
import os 
import subprocess
import sys

from struct import *
import intervaltree
import pysam

from collections import defaultdict

from modules import help_functions

def get_ultra_indexed_choordinates(ref_part_sequences, indexfolder, outfolder):
    id_to_chr = help_functions.pickle_load( os.path.join(indexfolder, 'id_to_chr.pickle') )
    indexed_regions = defaultdict(intervaltree.IntervalTree)

    for sequence_id, seq  in ref_part_sequences.items():
        chr_id, start, stop = unpack('LLL',sequence_id)
        ref_seq_name = id_to_chr[chr_id]
        indexed_regions[ref_seq_name].addi(start, stop, None)

    return indexed_regions

def align_with_minimap2(refs_path, read_path, outfolder, nr_cores, k_size):
    minimap2_samfile_path = os.path.join(outfolder, "minimap2.sam")
    with open(minimap2_samfile_path, "w") as output_file:
        # print('Running spoa...', end=' ')
        sys.stdout.flush()
        stderr_file = open(os.path.join(outfolder, "minimap2_errors.1") , 'w')
        # minimap2 --eqx -t 62 -ax splice -k13 -w 5 -G 500k {input.index} {input.fastq}
        subprocess.check_call([ 'minimap2',  '-ax', 'splice', '--eqx' , '-k', str(k_size), '-t', str(nr_cores), refs_path, read_path], stdout=output_file, stderr=stderr_file)
        # print('Done.')
        sys.stdout.flush()
    output_file.close()
    return minimap2_samfile_path

def get_exons_from_cigar(read):
    # https://pysam.readthedocs.io/en/latest/api.html
    # M   BAM_CMATCH  0
    # I   BAM_CINS    1
    # D   BAM_CDEL    2
    # N   BAM_CREF_SKIP   3
    # S   BAM_CSOFT_CLIP  4
    # H   BAM_CHARD_CLIP  5
    # P   BAM_CPAD    6
    # =   BAM_CEQUAL  7
    # X   BAM_CDIFF   8
    # B   BAM_CBACK   9

    aligned_choordinates = set()
    q_start = read.reference_start
    q_end = read.reference_end
    exon_sites = []
    ref_pos = q_start
    exon_start = q_start
    aln_cig_types = {2, 7, 0, 8} # 

    for i, (t, l) in enumerate(read.cigartuples):
        if t in aln_cig_types:
            ref_pos += l
        elif t == 3:
            exon_sites.append( (exon_start, ref_pos) )
            ref_pos += l
            exon_start = ref_pos

        # elif t == "I" or t == "S" or t == "H": # insertion or softclip
        #     ref_pos += 0
        # else: # reference skip or soft/hardclip "~", or match =
        #     print("UNEXPECTED!", t)
        #     sys.exit()

    exon_sites.append( (exon_start, ref_pos) )
    return exon_sites


def overlap_size(a, b, c, d):
    max_start = max(a, c)
    min_stop = min(b, d)
    return min_stop - max_start

def is_overlapping(a_start,a_stop, b_start,b_stop):
    return (int(a_start) <= int(b_start) <= int(a_stop) )  or (int(a_start) <= int(b_stop) <= int(a_stop)) or (int(b_start) <= int(a_start) <= int(a_stop) <= int(b_stop) )


def parse_alignments_and_mask(minimap2_samfile_path, indexed_regions, genomic_frac_cutoff):
    SAM_file = pysam.AlignmentFile(minimap2_samfile_path, "r", check_sq=False)
    reads_unindexed = {}
    reads_indexed = {}
    for read in SAM_file.fetch(until_eof=True):
        if read.flag == 0 or read.flag == 16:

            aligned_choordinates = get_exons_from_cigar(read)
            total_overlap = 0
            total_aligned_length = 0
            for (a_start, a_stop) in aligned_choordinates:
                overlaps = indexed_regions[read.reference_name].overlap(a_start, a_stop)
                for ovl in overlaps: #continue here to get sizxe of overlap
                    # print(ovl.begin, ovl.end)
                    total_overlap += overlap_size(a_start, a_stop, ovl.begin, ovl.end)
                # print("overlaps", overlaps)
                # total_overlap += overlaps
                total_aligned_length += a_stop - a_start

            # is_exonic = 1 if indexed_regions[reference_name].overlaps(reference_start, reference_end) else 0
            # print(read.query_name, "unindexed portion:", 1 - (total_overlap/total_aligned_length))
            if 1 - (total_overlap/total_aligned_length) > genomic_frac_cutoff:
                reads_unindexed[read.query_name] = read
            else:
                reads_indexed[read.query_name] = read

    return reads_unindexed, reads_indexed, SAM_file


def print_read_categories(reads_unindexed, reads_indexed, reads, outfolder, SAM_file):
    reads_to_align = open(os.path.join(outfolder, "reads_after_genomic_filtering.fasta"), "w")
    unindexed_aligned = pysam.AlignmentFile(os.path.join(outfolder, "unindexed.sam"), "w", template=SAM_file)
    indexed_aligned = pysam.AlignmentFile(os.path.join(outfolder, "indexed.sam"), "w", template=SAM_file)

    for acc, (seq, _) in help_functions.readfq(open(reads,"r")):
        if acc in reads_unindexed:
            read = reads_unindexed[acc]
            read.set_tag('XA', "")
            read.set_tag('XC', "uLTRA_unindexed")
            unindexed_aligned.write(read)
        else:
            reads_to_align.write(">{0}\n{1}\n".format(acc, seq))
            if acc in reads_indexed:
                read = reads_indexed[acc]
                indexed_aligned.write(read)

    unindexed_aligned.close()
    indexed_aligned.close()
    reads_to_align.close()
    return reads_to_align.name



def main(ref_part_sequences, ref, reads, outfolder, indexfolder, nr_cores, genomic_frac_cutoff, k_size):
    indexed_regions = get_ultra_indexed_choordinates(ref_part_sequences, indexfolder, outfolder)
    minimap2_samfile_path = align_with_minimap2(ref, reads, outfolder, nr_cores, k_size)
    reads_unindexed, reads_indexed, SAM_file = parse_alignments_and_mask(minimap2_samfile_path, indexed_regions, genomic_frac_cutoff)
    path_reads_to_align = print_read_categories(reads_unindexed, reads_indexed, reads, outfolder, SAM_file)
    return len(reads_unindexed), path_reads_to_align










from collections import defaultdict
from array import array
from struct import *
# import intervaltree

from modules.help_functions import readfq

def reverse_mapping(d):
    d_dict = defaultdict(list)
    for k,v in d.items():
        for i in v:
            d_dict[i].append(k)
    return d_dict

def dd_set(): # top level function declaration needed for multiprocessing
    return defaultdict(set)

def dd_tuple(): # top level function declaration needed for multiprocessing
    return defaultdict(tuple)

def add_items(array, chr_id, p1, p2):
    array.append(chr_id)
    array.append(p1)
    array.append(p2)

# def add_tiling(p1, p2, active_gene_ids, active_start, active_stop, min_segment_size, chr_id,
#                  tiling_segment_id_to_choordinates, tiling_segment_to_gene, tiling_segment_to_ref,
#                             tiling_parts_to_segments, tiling_gene_to_small_segments):
#     part_name = array("L", [chr_id, active_start, active_stop]).tobytes()
#     k = 0
#     while p2 - p1 - k > min_segment_size:
#         # tiling_segment_name = "segm_{0}_{1}_{2}".format(chr_id, p1+k, p1 + k + min_segment_size)
#         # tiling_segment_name = tiling_segment_id
#         tiling_segment_key = array("L", [chr_id, p1+k, p1 + k + min_segment_size]).tobytes()
#         tiling_segment_id_to_choordinates[tiling_segment_key] = (p1+k, p1 + k + min_segment_size)
#         tiling_segment_to_ref[tiling_segment_key] = chr_id
#         # tiling_parts_to_segments[part_name].append(tiling_segment_key)
#         add_items(tiling_parts_to_segments[part_name], chr_id, p1+k, p1 + k + min_segment_size)
#         tiling_segment_to_gene[tiling_segment_key] =  active_gene_ids 
#         # total_unique_segment_counter += min_segment_size
#         for gene_id in active_gene_ids:
#             # tiling_gene_to_small_segments[gene_id].append(tiling_segment_key)
#             add_items(tiling_gene_to_small_segments[gene_id], chr_id, p1+k, p1 + k + min_segment_size)
#         k += min_segment_size

#     # tiling_segment_name = "segm_{0}_{1}_{2}".format(chr_id, p1 + k - min_segment_size, p2)
#     # tiling_segment_name = tiling_segment_id
#     tiling_segment_key = array("L", [chr_id, max(0, p1 + k - min_segment_size), p2]).tobytes()
#     tiling_segment_id_to_choordinates[tiling_segment_key] = (max(0, p1 + k - min_segment_size), p2)
#     tiling_segment_to_ref[tiling_segment_key] = chr_id
#     # tiling_parts_to_segments[part_name].append(tiling_segment_key)
#     add_items(tiling_parts_to_segments[part_name], chr_id, max(0, p1 + k - min_segment_size), p2)
#     tiling_segment_to_gene[tiling_segment_key] =  active_gene_ids 
#     # total_unique_segment_counter += p2 - (p1 + k - min_segment_size)
#     for gene_id in active_gene_ids:
#         # tiling_gene_to_small_segments[gene_id].append(tiling_segment_key)
#         add_items(tiling_gene_to_small_segments[gene_id], chr_id, max(0, p1 + k - min_segment_size), p2)




def get_canonical_segments(part_to_canonical_pos, part_count_to_choord, part_to_active_gene, pos_to_exon_ids, exon_id_to_choordinates, small_segment_threshold, min_segment_size):
    # parts_to_exons, exon_to_gene, exon_id_to_choordinates,
    segment_id_to_choordinates = {}
    segment_to_gene = {} 
    segment_to_ref = {}
    parts_to_segments = defaultdict(lambda :array("L")) #defaultdict(dd_array)
    gene_to_small_segments = defaultdict(lambda :array("L")) #defaultdict(lambda :array("b"))
    total_unique_segment_counter = 0
    total_segments_bad = 0
    bad = 0

    # for all the poorly fitting reads to annotated start stop exon sites we employ a tiling segment structure and 
    # align reads in a second pass
    # M2 = defaultdict(lambda: defaultdict(lambda :array("I")))
    tiling_segment_id_to_choordinates = {}
    tiling_segment_to_gene = {} 
    tiling_segment_to_ref = {}
    tiling_parts_to_segments = defaultdict(lambda :array("L")) #defaultdict(dd_array)
    tiling_gene_to_small_segments = defaultdict(lambda :array("L")) #= defaultdict(lambda :array("b"))


    # tiling_structures = [tiling_segment_id_to_choordinates, tiling_segment_to_gene, tiling_segment_to_ref, tiling_parts_to_segments, tiling_gene_to_small_segments]

    for (chr_id, part_id) in part_to_canonical_pos:
        active_start, active_stop = part_count_to_choord[(chr_id, part_id)]
        part_name = array("L", [chr_id, active_start, active_stop]).tobytes()
        active_gene_ids = part_to_active_gene[(chr_id, part_id)]
        sorted_pos = sorted(part_to_canonical_pos[(chr_id, part_id)])

        open_starts_e_ids = set() #list( pos_to_exon_ids[(chr_id, part_id)][p, is_start] for p, is_start in pos_to_exon_ids[(chr_id, part_id)] if p < ) # exons spanning over the small segment
        pos_tuples = [(p1, p2) for p1, p2 in zip(sorted_pos[:-1], sorted_pos[1:])]
        for i, (p1, p2) in enumerate(pos_tuples):
            open_starts_e_ids.update(pos_to_exon_ids[(chr_id, part_id)][p1, True]) # add the exons that start at this point 
            open_starts_e_ids.difference_update(pos_to_exon_ids[(chr_id, part_id)][p1, False]) # remove the ones that ended here
            # add_tiling(p1, p2, active_gene_ids, active_start, active_stop, min_segment_size, chr_id,
            #             tiling_segment_id_to_choordinates, tiling_segment_to_gene, tiling_segment_to_ref,
            #             tiling_parts_to_segments, tiling_gene_to_small_segments)
            if p2 - p1 >= min_segment_size:
                # print("here good", p2 - p1)
                # segment_name = "segm_{0}_{1}_{2}".format(chr_id,p1,p2)
                # segment_name = segment_id
                segment_name = array("L", [chr_id, p1, p2]).tobytes()
                # temp_already_added_choord.add((chr_id, p1, p2))
                segment_id_to_choordinates[segment_name] = (p1, p2)
                segment_to_ref[segment_name] = chr_id
                # parts_to_segments[part_name].append(segment_name)
                add_items(parts_to_segments[part_name], chr_id, p1, p2)
                segment_to_gene[segment_name] =  active_gene_ids 
                total_unique_segment_counter += p2 - p1
                # print("Std:", segment_name, (chr_id, p1, p2))
                open_starts_e_ids.difference_update(pos_to_exon_ids[(chr_id, part_id)][p2, False])
                # add small segments
                if p2 - p1 <= small_segment_threshold:
                    for gene_id in active_gene_ids:
                        # gene_to_small_segments[gene_id].append(segment_name)
                        add_items(gene_to_small_segments[gene_id], chr_id, p1, p2)
                # segment_id += 1

            else:
                # add_earlier = False
                if i > 0: 
                    k = 1
                    while i - k >= 0:
                        if p2 - pos_tuples[i-k][0] >= min_segment_size: # or i - k == 0: # last case is i - k == 0 and we still have not added anything
                            # segment_name = "segm_{0}_{1}_{2}".format(chr_id, pos_tuples[i-k][0],p2)
                            # exon_name = "exon_{0}_{1}_{2}".format(chr_id, pos_tuples[i-k][0], p2) 
                            # segment_name = segment_id
                            segment_name = array("L", [chr_id, pos_tuples[i-k][0], p2]).tobytes()

                            if segment_name not in segment_id_to_choordinates: #(chr_id, pos_tuples[i-k][0],p2) not in temp_already_added_choord: #segment_id_to_choordinates and exon_name not in segment_id_to_choordinates:
                                # print("Added 1:", k, p2 - p1, p2 - pos_tuples[i-k][0])
                                total_segments_bad += p2 - pos_tuples[i-k][0]

                                segment_id_to_choordinates[segment_name] = (pos_tuples[i-k][0], p2)
                                segment_to_ref[segment_name] = chr_id
                                # parts_to_segments[part_name].append(segment_name)
                                add_items(parts_to_segments[part_name], chr_id, pos_tuples[i-k][0], p2)
                                segment_to_gene[segment_name] =  active_gene_ids 
                                # temp_already_added_choord.add( (chr_id, pos_tuples[i-k][0],p2) )
                                # print("while before:", segment_name,(chr_id, pos_tuples[i-k][0],p2))
                                # segment_id += 1

                                # add_earlier = True
                                # add small segments
                                if p2 - pos_tuples[i-k][0] <= small_segment_threshold:
                                    for gene_id in active_gene_ids:
                                        # gene_to_small_segments[gene_id].append(segment_name)
                                        add_items(gene_to_small_segments[gene_id], chr_id, pos_tuples[i-k][0], p2)
                            break
                        else:
                            pass
                            # print("DID not make it 1:", k, p2 - pos_tuples[i-k][0], p1, p2)

                        k += 1

                # add_later = False
                if i < len(pos_tuples) - 1:
                    k = 1
                    while i + k <= len(pos_tuples) - 1: 
                        if pos_tuples[i+k][1] - p1 >= min_segment_size: # or i + k == len(pos_tuples) - 1:  # last case is i - k == 0 and we still have not added anything
                            # segment_name = "segm_{0}_{1}_{2}".format(chr_id,p1,pos_tuples[i+k][1])
                            # exon_name = "exon_{0}_{1}_{2}".format(chr_id,p1,pos_tuples[i+k][1]) 
                            # segment_name = segment_id
                            segment_name = array("L", [chr_id, p1, pos_tuples[i+k][1]]).tobytes()

                            if (chr_id, p1, pos_tuples[i+k][1]) not in segment_id_to_choordinates: # not in temp_already_added_choord: # segment_name not in segment_id_to_choordinates and exon_name not in segment_id_to_choordinates: 
                                # print("Added 2:", k, p2 - p1, pos_tuples[i+k][1] - p1)
                                total_segments_bad += pos_tuples[i+k][1] - p1

                                segment_id_to_choordinates[segment_name] = (p1, pos_tuples[i+k][1])
                                segment_to_ref[segment_name] = chr_id
                                # parts_to_segments[part_name].append(segment_name)
                                add_items(parts_to_segments[part_name], chr_id, p1, pos_tuples[i+k][1])
                                segment_to_gene[segment_name] =  active_gene_ids 
                                # temp_already_added_choord.add( (chr_id, p1, pos_tuples[i+k][1]) )
                                # print("while after:", segment_name, (chr_id, p1, pos_tuples[i+k][1]))

                                # segment_id += 1

                                # add_later = True
                                # add small segments
                                if pos_tuples[i+k][1] - p1 <= small_segment_threshold:
                                    for gene_id in active_gene_ids:
                                        # gene_to_small_segments[gene_id].append(segment_name)
                                        add_items(gene_to_small_segments[gene_id], chr_id, p1, pos_tuples[i+k][1])
                            break
                        else:
                            pass
                            # print("DID not make it 2:", k, pos_tuples[i+k][1] - p1, p1, p2)
                        k += 1

                if len(pos_tuples) == 1:
                    # segment_name = "segm_{0}_{1}_{2}".format(chr_id,p1,p2)
                    # segment_name = segment_id
                    segment_name = array("L", [chr_id, p1, p2]).tobytes()

                    segment_id_to_choordinates[segment_name] = (p1, p2)
                    segment_to_ref[segment_name] = chr_id
                    # parts_to_segments[part_name].append(segment_name)
                    add_items(parts_to_segments[part_name], chr_id, p1, p2)
                    segment_to_gene[segment_name] =  active_gene_ids 
                    total_unique_segment_counter += p2 - p1
                    # temp_already_added_choord.add( (chr_id, p1, p2) )
                    # print("Third:", segment_name,  (chr_id, p1, p2))
                    # segment_id += 1

                    # add small segments
                    if p2 - p1 <= small_segment_threshold:
                        for gene_id in active_gene_ids:
                            # gene_to_small_segments[gene_id].append(segment_name)
                            add_items(gene_to_small_segments[gene_id], chr_id, p1, p2)


                # either part consists only of a small exon (e7 in fig in paper) or 
                # one of the adjacent segment cutpoints (immediately before and/or after) are too close
                # which also makes them not pass the minimum segment size
                # in this case simply add the exons
                # if not add_later or not add_earlier: 
                relevant_starts = list(pos_to_exon_ids[(chr_id, part_id)][p1, True])
                relevant_ends = list(pos_to_exon_ids[(chr_id, part_id)][p2, False])
                all_segm_spanning = set(relevant_starts + relevant_ends)
                if not all_segm_spanning:
                    all_segm_spanning = open_starts_e_ids
                bad += p2 - p1
                for e_id in all_segm_spanning:
                    e_start, e_stop = exon_id_to_choordinates[e_id]
                    # add_tiling(e_start, e_stop, active_gene_ids, active_start, active_stop, min_segment_size, chr_id,
                    #             tiling_segment_id_to_choordinates, tiling_segment_to_gene, tiling_segment_to_ref,
                    #             tiling_parts_to_segments, tiling_gene_to_small_segments)

                    # exon_name = "exon_{0}_{1}_{2}".format(chr_id,e_start,e_stop) 
                    # segment_name = segment_id
                    segment_name = array("L", [chr_id, e_start, e_stop]).tobytes()

                    if segment_name not in segment_id_to_choordinates: #(chr_id, e_start, e_stop) not in temp_already_added_choord: #"segm_{0}_{1}_{2}".format(chr_id,e_start,e_stop) not in segment_id_to_choordinates:                
                        segment_id_to_choordinates[segment_name] = (e_start, e_stop) 
                        segment_to_ref[segment_name] = chr_id
                        # parts_to_segments[part_name].append(segment_name)
                        add_items(parts_to_segments[part_name], chr_id, e_start, e_stop)
                        segment_to_gene[segment_name] = active_gene_ids
                        total_segments_bad += e_stop - e_start
                        # temp_already_added_choord.add( (chr_id, e_start, e_stop) )
                        # print("Fourth:", segment_name,  (chr_id, e_start, e_stop))

                        # segment_id += 1

                        if e_stop - e_start <= small_segment_threshold:
                            # gene_to_small_segments[gene_id].append(segment_name)
                            add_items(gene_to_small_segments[gene_id], chr_id, e_start, e_stop)
                    else:
                        pass
                        # print("LOOOOOOL", e_stop - e_start)


                open_starts_e_ids.difference_update(pos_to_exon_ids[(chr_id, part_id)][p2, False])


    print("total_unique_segment_counter", total_unique_segment_counter)
    print("total_segments_bad", total_segments_bad)
    print("bad", bad)
    return parts_to_segments, segment_to_gene, segment_id_to_choordinates, segment_to_ref, gene_to_small_segments 
    # return parts_to_segments, segment_to_gene, segment_id_to_choordinates, segment_to_ref, gene_to_small_segments, tiling_structures 


def add_to_chr_mapping(chr_name, chr_to_id, id_to_chr):
    hash_id = len(chr_to_id)
    if chr_name in chr_to_id:
        return chr_to_id[chr_name]
    else:
        hash_id += 1
        chr_to_id[chr_name] = hash_id
        id_to_chr[hash_id] = chr_name
        return hash_id

        
def create_graph_from_exon_parts(db, flank_size, small_exon_threshold, min_segment_size, refs_lengths): 
    """
        We need to link parts --> exons and exons --> transcripts
    """
    # print(dir(db))
    # print([c for c in db.features_of_type('chr', order_by='seqid')])
    # print( [c for c in db.all_features()])
    # sys.exit()
    chr_to_id = {}
    id_to_chr = {}
    # genes_to_ref = {} # gene_id : { (exon_start, exon_stop) : set() }
    # exons_to_ref = {} # gene_id : { (exon_start, exon_stop) : set() }
    # exon_to_gene = {} # exon_id : [gene_id ]

    flank_ids = set()  
    total_flanks2 = 0
    total_flank_size = 0

    exon_id_to_choordinates = {}
    splices_to_transcripts = defaultdict(dd_set)
    all_splice_pairs_annotations = defaultdict(dd_set)
    all_splice_sites_annotations = defaultdict(set)
    # annotated_transcripts = defaultdict(set)
    # part_intervals = defaultdict(intervaltree.IntervalTree)
    parts_to_exons = defaultdict(dd_set)
    max_intron_chr = defaultdict(int)
    pos_to_exon_ids = defaultdict(dd_set)
    part_counter = 0
    part_to_canonical_pos = defaultdict(set)
    part_count_to_choord = defaultdict(dd_tuple)
    part_to_active_gene = defaultdict(set)
    exon_ids = set() #defaultdict(dd_set)

    for i, exon in enumerate(db.features_of_type('exon', order_by='seqid')):
        chr_name = exon.seqid
        # if chr_name in refs_lengths:
        #     chr_length = refs_lengths[chr_name]
        #     if exon.start - 1 < 0 or exon.stop > chr_length:
        #         print("Warning. Annotation on chromosome {0} outside chromosome length [0,{1}]. Annotation was: ({2}, {3})".format(chr_name,chr_length,exon.start - 1, exon.stop))
        # if chr_name == "211000022278436":
        #     print("chromosome length [0,{1}]. Annotation was: ({2}, {3})".format(chr_name,chr_length,exon.start - 1, exon.stop))
        chr_id = add_to_chr_mapping(chr_name, chr_to_id, id_to_chr)
        # exons_to_ref[exon.id] = chr_id
        # print(dir(exon))
        # print(exon.attributes["gene_id"])
        exon_gene_ids = exon.attributes["gene_id"] # is a list of strings
        # exon_to_gene[exon.id] = exon_gene_ids
        exon_id_to_choordinates[exon.id] = (exon.start - 1, exon.stop)
        exon_name = array("L", [chr_id, exon.start - 1, exon.stop]).tobytes()
        exon_ids.add(exon_name)
        # creating the augmentation
        if i == 0: # initialization
            prev_chr_id = chr_id
            prev_chr_name = chr_name
            active_start = exon.start - 1
            active_stop = exon.stop
            active_exons = set() 
            active_exons.add(exon.id)    
            active_gene_ids = set(exon_gene_ids) 
            # adding very first flank on chromosome
            # print(max(0, exon.start - 1000), exon.start - 1)
            # to avoid flank intervals of length 0
            if exon.start - 1 > 0:
                flank_name =  array("L", [chr_id, max(0, exon.start - 2*flank_size), exon.start - 1]).tobytes()
                flank_ids.add(flank_name)
                total_flanks2 += 1
                total_flank_size += 2*flank_size

            part_to_canonical_pos[(chr_id, part_counter)].add(exon.start - 1)
            part_to_canonical_pos[(chr_id, part_counter)].add(exon.stop)
            part_to_active_gene[(chr_id, part_counter)].update(set(exon_gene_ids))

            continue
            # print(2*flank_size)

        if chr_id != prev_chr_id: # switching chromosomes
            parts_to_exons[prev_chr_id][(active_start, active_stop)] = active_exons
            part_count_to_choord[(prev_chr_id,part_counter)] = (active_start, active_stop)
           
            # adding the very last flank on previous chromosome
            # chr_length = refs_lengths[prev_chr_name]
            chr_length = refs_lengths.get(prev_chr_name, active_stop + 2*flank_size) # if key 'prev_chr_name' is not present in refs_lengths (i.e. in reference fasta), it means that we will not use the annotations to this chromosome anyway so value does not matter..
            # to avoid flank intervals of length 0
            if active_stop < chr_length:
                flank_name = array("L", [prev_chr_id, max(0, active_stop), min(chr_length, active_stop + 2*flank_size)]).tobytes()
                flank_ids.add(flank_name)
                total_flanks2 += 1
                total_flank_size += 2*flank_size
                # print(2*flank_size)

            # print(max(0, active_stop), active_stop + 1000)
            prev_chr_id = chr_id
            prev_chr_name = chr_name
            active_start = exon.start - 1
            active_stop = exon.stop

            # adding the very first flank on new chromosome
            # to avoid flank intervals of length 0
            if exon.start - 1 > 0:
                flank_name = array("L", [chr_id, max(0, exon.start - 2*flank_size), exon.start - 1]).tobytes()
                flank_ids.add(flank_name)
                total_flanks2 += 1
                total_flank_size += 2*flank_size

            part_counter +=1
            part_to_canonical_pos[(chr_id, part_counter)].add(exon.start - 1)
            part_to_canonical_pos[(chr_id, part_counter)].add(exon.stop)
            part_to_active_gene[(chr_id, part_counter)].update(set(exon_gene_ids))
            pos_to_exon_ids[(chr_id, part_counter)][(exon.start - 1, True)].add(exon.id)
            pos_to_exon_ids[(chr_id, part_counter)][ (exon.stop, False)].add(exon.id) 

            active_exons = set() 
            active_exons.add(exon.id)     
            active_gene_ids = set(exon_gene_ids) 


        elif exon.start - 1 > active_stop + 20:
            parts_to_exons[chr_id][(active_start, active_stop)] = active_exons
            part_count_to_choord[(chr_id,part_counter)] = (active_start, active_stop)
            segment_size = 2*flank_size if set(exon_gene_ids).isdisjoint(active_gene_ids) else flank_size
            # part_intervalst(segment_size)
            if exon.start - active_stop > 2*segment_size:
                # chr_length = refs_lengths[chr_name]
                chr_length = refs_lengths.get(chr_name, active_stop + segment_size) # if key 'chr_name' is not present in refs_lengths (i.e. in reference fasta), it means that we will not use the annotations to this chromosome anyway so value does not matter..
                flank_name = array("L", [chr_id, max(0, active_stop), min(chr_length, active_stop + segment_size)]).tobytes()
                flank_ids.add(flank_name)
                flank_name = array("L", [chr_id, max(0, exon.start - segment_size), exon.start - 1]).tobytes()
                flank_ids.add(flank_name) 
                total_flanks2 += 2
                total_flank_size += 2*segment_size
                # print(max(0, active_stop), active_stop + 1000)
                # print(max(0, exon.start - 1000), exon.start - 1)

            else: # add the whole intron
                flank_name = array("L", [chr_id, max(0, active_stop), exon.start - 1]).tobytes()
                flank_ids.add(flank_name)
                total_flanks2 += 1
                total_flank_size += (exon.start - 1 - max(0, active_stop))

            active_exons = set()
            active_exons.add(exon.id)
            active_gene_ids = set(exon_gene_ids) 

            active_start = exon.start - 1
            active_stop = exon.stop
            part_counter +=1
            part_to_canonical_pos[(chr_id, part_counter)].add(exon.start - 1)
            part_to_canonical_pos[(chr_id, part_counter)].add(exon.stop)
            part_to_active_gene[(chr_id, part_counter)].update(set(exon_gene_ids))
            pos_to_exon_ids[(chr_id, part_counter)][(exon.start - 1, True)].add(exon.id)
            pos_to_exon_ids[(chr_id, part_counter)][ (exon.stop, False)].add(exon.id)

        else:
            active_exons.add(exon.id)    
            active_stop = max(active_stop, exon.stop)
            active_gene_ids.update(exon_gene_ids) 
            part_to_canonical_pos[(chr_id, part_counter)].add(exon.start - 1)
            part_to_canonical_pos[(chr_id, part_counter)].add(exon.stop)
            part_to_active_gene[(chr_id, part_counter)].update(exon_gene_ids)
            pos_to_exon_ids[(chr_id, part_counter)][(exon.start - 1, True)].add(exon.id)
            pos_to_exon_ids[(chr_id, part_counter)][ (exon.stop, False)].add(exon.id)

        assert active_start <= exon.start - 1

    if i > 0:
        # addig the very last flank at the last chromosome in the annotation
        # chr_length = refs_lengths[chr_name]
        chr_length = refs_lengths.get(chr_name, active_stop + 2*flank_size) # if key 'chr_name' is not present in refs_lengths (i.e. in reference fasta), it means that we will not use the annotations to this chromosome anyway so value does not matter..
    else:
        print("No record was read from the GTF file! This is likely because of a misformatting issue. For example, check so that 'exon', 'transcript', and 'gene' are found in column 3.")
        print("Stopping here.")
        sys.exit()

    # to avoid flank intervals of length 0
    if active_stop < chr_length:
        flank_name = array("L", [chr_id, max(0, active_stop), min(chr_length, active_stop + 2*flank_size)]).tobytes()
        flank_ids.add(flank_name)
        total_flanks2 += 1
        total_flank_size += 2*flank_size

    # print("NR EXONS + COMPL:", len(exon_to_gene))
    print("total_flanks2:", total_flanks2)
    print("total_flank_size", total_flank_size)
    # print(flank_ids)
    parts_to_exons[chr_id][(active_start, active_stop)] = active_exons
    part_count_to_choord[(chr_id,part_counter)] = (active_start, active_stop)

    # parts_to_segments, segment_to_gene, \
    # segment_id_to_choordinates, segment_to_ref, \
    # gene_to_small_segments, tiling_structures  = get_canonical_segments(part_to_canonical_pos, part_count_to_choord, part_to_active_gene, pos_to_exon_ids, exon_id_to_choordinates, small_exon_threshold, min_segment_size)

    parts_to_segments, segment_to_gene, \
    segment_id_to_choordinates, segment_to_ref, \
    gene_to_small_segments  = get_canonical_segments(part_to_canonical_pos, part_count_to_choord, part_to_active_gene, pos_to_exon_ids, exon_id_to_choordinates, small_exon_threshold, min_segment_size)


    print("total parts size:", sum( [stop - start for chrrr in parts_to_exons for start,stop in parts_to_exons[chrrr] ]))
    print("total exons size:", sum( [stop - start for start, stop in exon_id_to_choordinates.values() ]))
    # print(chr_to_id)
    # print(id_to_chr)

    min_intron = 2**32

    for transcript in db.features_of_type('transcript', order_by='seqid'): #db.children(gene, featuretype='transcript', order_by='start'):
        # if transcript.seqid.isdigit() or transcript.seqid == 'X' or transcript.seqid == 'Y':
        #     chr_name = 'chr'+ transcript.seqid
        # elif transcript.seqid == 'MT':
        #     chr_name = 'chrM'
        # else:
        chr_name = transcript.seqid
        chr_id = chr_to_id[chr_name]
        # print("here", transcript.id,  str(chr_id))  
        transcript_exons = []
        for exon in db.children(transcript, featuretype='exon', order_by='start'):
            transcript_exons.append( (exon.start-1, exon.stop) )
        internal_transcript_splices = [ (e1[1],e2[0]) for e1, e2 in zip(transcript_exons[:-1],transcript_exons[1:])]
        
        for i1,i2 in internal_transcript_splices:
            if i2 - i1 > max_intron_chr[chr_id]:
                max_intron_chr[chr_id] = i2 - i1
                # print("intron size:", i2 - i1, chr_id)

            if i2 - i1 < min_intron:
                min_intron = i2 - i1
        # internal transcript splices
        splices_to_transcripts[chr_id][ tuple(internal_transcript_splices)].add(transcript.id)
        for site1, site2 in internal_transcript_splices:
            all_splice_pairs_annotations[chr_id][(site1, site2)].add( transcript.id )
            # if site2 == 3105:
            #     print('LOOOL',chr_id)
            #     ccc.add(chr_id)

            all_splice_sites_annotations[chr_id].add(site1)
            all_splice_sites_annotations[chr_id].add(site2)
        
        # add start and end splice to all_splice_sites_annotations 
        if transcript_exons:
            all_splice_sites_annotations[chr_id].add(transcript_exons[0][0])
            all_splice_sites_annotations[chr_id].add(transcript_exons[-1][-1])
        else:
            print("Something is wrong with transcript annotation: {0} on gene: {1}, and could not be added. Check so that the gene ID and transcript ID are not the same!".format(transcript.id, transcript.seqid))
            # sys.exit()
    print("min_intron:", min_intron)
    # sys.exit()
    # transcripts_to_splices = reverse_mapping(splices_to_transcripts)
    transcripts_to_splices = defaultdict(dict)
    for chr_id, chr_sp_sites_dict in splices_to_transcripts.items():
        for unique_sp_sites, tr_ids in chr_sp_sites_dict.items():
            for tr_id in tr_ids:
                transcripts_to_splices[chr_id][tr_id] = unique_sp_sites

    # print(chr_to_id)
    # for f in flank_ids:
    #     chr_id, start, stop = unpack('LLL',f)
    #     if chr_id == 7:
    #         print(chr_id, start, stop)

    return  segment_to_ref, parts_to_segments, splices_to_transcripts, \
            transcripts_to_splices, all_splice_pairs_annotations, \
            all_splice_sites_annotations, segment_id_to_choordinates, \
            segment_to_gene, gene_to_small_segments, flank_ids, max_intron_chr, \
            exon_ids, chr_to_id, id_to_chr

    # return  segment_to_ref, parts_to_segments, splices_to_transcripts, \
    #         transcripts_to_splices, all_splice_pairs_annotations, \
    #         all_splice_sites_annotations, segment_id_to_choordinates, \
    #         segment_to_gene, gene_to_small_segments, flank_ids, max_intron_chr, \
    #         exon_ids, chr_to_id, id_to_chr, tiling_structures

def get_sequences_from_choordinates(sequence_choordinates, refs):
    sequence_container = defaultdict(dict)
    for sequence_id in sequence_choordinates:
        chr_id, start, stop = unpack('LLL',sequence_id)
        # start,stop = sequence_choordinates[sequence_id]
        # chr_id = segments_to_ref[sequence_id]
        if chr_id not in refs:
            continue
        else:
            seq = refs[chr_id][start : stop] 
            # key = array("L", [chr_id,start,stop]).tobytes()
            sequence_container[sequence_id] = seq
    # print(segments)
    return sequence_container


# Functions for masking overly abundant kmers 

import itertools
from collections import defaultdict, deque
def kmer_counter(ref_part_sequences, kmer_size):
    count = defaultdict(int)
    position_count = defaultdict(list)
    it_count = 0
    for sequence_id, seq  in ref_part_sequences.items():
        # for part, seq in ref_part_sequences[chr_id].items():
        read_kmers = deque([seq[i:i+kmer_size] for i in range(len(seq) - kmer_size + 1)])
        for i, kmer in enumerate(read_kmers):
            count[kmer] += 1
            it_count += 1

            if it_count % 10000000 == 0:
                print(int(it_count/10000000.0), " million kmers processed.")
                # clean every 1M kmers
                if it_count % 20000000 == 0:
                    for kmer in list(count.keys()):
                        if count[kmer] == 1:
                            del count[kmer]
                    print(len(count),"kmers stored")
    return count


def mask_refs(ref_part_sequences, to_mask, kmer_size):
    mask_counter = 0
    tot_counter = 0
    for sequence_id, seq in ref_part_sequences.items():
        # chr_id, start, stop = unpack('LLL',sequence_id)
        # seq = ref_part_sequences[sequence_id]
        # for part, seq in ref_part_sequences[chr_id].items():
        tot_counter += 1
        read_kmers = [seq[i:i+kmer_size] for i in range(len(seq) - kmer_size + 1 )]
        seq_masked = []
        has_been_modified = False
        if read_kmers:
            # separately treat first kmer in every seq
            if read_kmers[0] in to_mask:
                seq_masked.append(read_kmers[0][:-1]+'N')
                has_been_modified = True
            else:
                seq_masked.append(read_kmers[0])

            for i, kmer in enumerate(read_kmers[1:]):
                if kmer in to_mask:
                    seq_masked.append("N") # mask the last base
                    has_been_modified = True
                else:
                    seq_masked.append(kmer[-1]) # add the true last base
            
            if has_been_modified:
                seq_masked = "".join([s for s in seq_masked])
                # print("masking", seq, "to", seq_masked) 
                ref_part_sequences[sequence_id] = seq_masked
                mask_counter += 1
    print(mask_counter, "{0} out of {1} sequences has been modified in masking step.".format(mask_counter, tot_counter))

def mask_abundant_kmers(ref_part_sequences, kmer_size, mask_threshold):
    DBG = kmer_counter(ref_part_sequences, kmer_size)
    n = float(len(DBG))
    print(n, "Unique kmers in reference part sequences with abundance > 1")
    to_mask = set()
    for i, (kmer, abundance) in enumerate(sorted(DBG.items(), key=lambda x: x[1], reverse=True)):
        if abundance >= mask_threshold:
            to_mask.add(kmer) 
            print(kmer, abundance)
        else:
            break

    mask_refs(ref_part_sequences, to_mask, kmer_size)
    # print(len(to_mask), "kemrs masked.")
    # print(sorted([DBG[kmer] for kmer in DBG], reverse=True))








import os
import subprocess
from sys import stdout, exit

from collections import defaultdict
from collections import namedtuple

mem = namedtuple('Mem', ['x', 'y', 'c', 'd', 'val', 'j', "exon_part_id"])
globals()[mem.__name__] = mem # Global needed for multiprocessing

def find_mems_mummer(outfolder, read_path, refs_path, mummer_out_path, min_mem):
    with open(mummer_out_path, "w") as output_file:
        # print('Running spoa...', end=' ')
        stdout.flush()
        null = open(os.path.join(outfolder, "mummer_errors.1") , "w")
        subprocess.check_call([ 'mummer',   '-maxmatch', '-l' , str(min_mem),  refs_path, read_path], stdout=output_file, stderr=null)
        # print('Done.')
        stdout.flush()
    output_file.close()


def find_mems_slamem(outfolder, read_path, refs_path, out_path, min_mem):
    # time slaMEM -l 14 /Users/kxs624/tmp/ULTRA/human_test/refs_sequences.fa /Users/kxs624/tmp/ULTRA/human_test_new_flanking_strat/reads_tmp.fq -o /Users/kxs624/tmp/ULTRA/human_test/slamem_test.tx
    # with open(out_path, "w") as output_file:
    tmp = out_path.split("seeds_batch_")[1]
    batch_id =  tmp.split('.')[0]
    stdout.flush()
    stderr_file = open(os.path.join(outfolder, "slamem_stderr_{0}.1".format(batch_id)) , "w")
    stdout_file = open(os.path.join(outfolder, "slamem_stdout_{0}.1".format(batch_id)) , "w")
    try: # slaMEM throws error if no MEMs are found in any of the sequences
        subprocess.check_call([ 'slaMEM', '-l' , str(min_mem),  refs_path, read_path, '-o', out_path ], stdout=stdout_file, stderr=stderr_file)
        print("Using SLAMEM")
    except:
        find_mems_mummer(outfolder, read_path, refs_path, out_path, min_mem)
        print("Using MUMMER")

    # print('Done.')
    stdout.flush()
    # output_file.close()


def find_nams_strobemap(outfolder, read_path, refs_path, out_path, nr_cores, min_mem):
    # /usr/bin/time -l ./StrobeMap -n 2 -k 9 -w 30 -t 3 -s  -o /Users/kxs624/tmp/ULTRA/human_test/refs_sequences.fa /Users/kxs624/tmp/ULTRA/human_test_new_flanking_strat/reads_tmp.fq
    # StrobeMap -n 2 -k 9 -w 30 -t 3 -s  -o ~/tmp/STROBEMERS/multithreading/ /Users/kxs624/tmp/ULTRA/dros_tmp/refs_sequences.fa /Users/kxs624/tmp/ULTRA/dros_test/reads_16xrep.fa
    # with open(out_path, "w") as output_file:
    stdout.flush()
    stderr_file = open(os.path.join(outfolder, "strobemap_stderr.1") , "w")
    stdout_file = open(os.path.join(outfolder, "strobemap_stdout.1") , "w")
    try: # slaMEM throws error if no MEMs are found in any of the sequences
        subprocess.check_call([ 'StrobeMap', '-n' , "2", '-k' , '10', '-v' , "11", '-w' , "35", '-C' , '500', '-L' , '1000', '-S', '-t', str(nr_cores), '-s', '-o', outfolder, refs_path, read_path ], stdout=stdout_file, stderr=stderr_file)
        print("Using StrobeMap")
        print([ 'StrobeMap', '-n' , "2", '-k' , '10', '-v' , "11", '-w' , "35", '-C' , '500', '-L' , '1000', '-S', '-t', str(nr_cores), '-s', '-o', outfolder, refs_path, read_path ])
    except:
        find_mems_slamem(outfolder, read_path, refs_path, out_path, min_mem)
        print("An unexpected error happend in StrobeMap, check error log at:", stderr_file)
        print("If you beileive this is a bug in StrobeMap, report an issue at: https://github.com/ksahlin/strobemers")
        print("You can always sidestep this issue by providing another seed finder to uLTRA, i.e., remove option --use_NAM_seeds.")
        sys.exit()
    stdout.flush()

# def parse_results(mems_path):
#     # file = open(os.path.join(mems_folder, "mummer_mems.txt"), 'r')
#     mems_db = {}
#     # read_mems_tmp = {}

#     for i, line in enumerate(open(mems_path, 'r')):
#         if line[0] == '>':
#             if i == 0:
#                 read_acc = line.split()[1].strip()  # mems_db[line.split()[1].strip)()] = [] 
#             else:
#                 mems_db[read_acc] = read_mems_tmp 
#                 read_acc = line.split()[1].strip() 
            
#             read_mems_tmp = defaultdict(list)

#         else:
#             vals =  line.split() #11404_11606           1     11405       202
#             exon_part_id = vals[0]
#             chr_id, ref_coord_start, ref_coord_end = exon_part_id.split('^')
#             mem_len = int(vals[3])
#             mem_ref_exon_part_start = int(vals[1])
#             mem_read_start = int(vals[2])
#             # convert to 0-indexed as python, however last coordinate is inclusive of the hit, not as in python end-indexing
#             mem_tuple = mem(int(ref_coord_start) + mem_ref_exon_part_start - 1, int(ref_coord_start) + mem_ref_exon_part_start -1 + mem_len - 1,
#                             mem_read_start-1, mem_read_start-1 + mem_len - 1, 
#                             mem_len, None, exon_part_id)
            
#             read_mems_tmp[chr_id].append( mem_tuple )
#         # print(line)
#     # add last read
#     mems_db[read_acc] = read_mems_tmp 


#     return mems_db


def get_mem_records(mems_path, reads):
    '''
        Reads contains all the relevant reads in the batch to read mems from 
    '''
    relevant = False
    relevant_read_cnt = 0
    for i, line in enumerate(open(mems_path, 'r')):
        if line[0] == '>':
            acc = line[1:].strip()
            if acc not in reads:
                relevant = False
                continue
            else:
                relevant = True
                relevant_read_cnt +=1

            if relevant_read_cnt == 1:
                read_acc = acc  
            else:

                for chr_id in list(read_mems_tmp.keys()):
                    coordinate_sorted_tuples = sorted(read_mems_tmp[chr_id], key = lambda x: x[1])
                    sorted_mems = [ mem(x,y,c,d,val,j,e_id) for j, (x, y, c, d, val, e_id) in enumerate(coordinate_sorted_tuples) ]
                    read_mems_tmp[chr_id] = sorted_mems

                yield read_acc, read_mems_tmp
                read_acc = acc 
            
            read_mems_tmp = defaultdict(list)

        elif relevant:
                vals =  line.split() #11404_11606           1     11405       202
                exon_part_id = vals[0]
                chr_id, ref_coord_start, ref_coord_end = exon_part_id.split('^')
                chr_id = int(chr_id)
                mem_len = int(vals[3])
                # convert to 0-indexed reference as in python
                # however, for MEM length last coordinate is inclusive of the hit in MEM solvers, not as in python end-indexing
                mem_ref_exon_part_start = int(vals[1]) - 1
                mem_read_start = int(vals[2]) - 1
                ref_coord_start = int(ref_coord_start) # has already been 0-indexed when constructing parts
                mem_genome_start = ref_coord_start + mem_ref_exon_part_start
                
                
                # mem_tuple = mem(int(ref_coord_start) - 1 + mem_ref_exon_part_start - 1, int(ref_coord_start) - 1 + mem_ref_exon_part_start -1 + mem_len - 1,
                #                 mem_read_start-1, mem_read_start-1 + mem_len - 1, 
                #                 mem_len, None, exon_part_id)
                # read_mems_tmp[chr_id].append( mem_tuple )
                info_tuple = ( mem_genome_start, mem_genome_start + mem_len - 1,
                                mem_read_start, mem_read_start + mem_len - 1, 
                                mem_len, exon_part_id)
                read_mems_tmp[chr_id].append( info_tuple )


    for chr_id in list(read_mems_tmp.keys()):
        coordinate_sorted_tuples = sorted(read_mems_tmp[chr_id], key = lambda x: x[1])
        sorted_mems = [ mem(x,y,c,d,val,j,e_id) for j, (x, y, c, d, val, e_id) in enumerate(coordinate_sorted_tuples) ]
        read_mems_tmp[chr_id] = sorted_mems
    print("READ {0} RECORDS.".format(relevant_read_cnt))
    yield read_acc, read_mems_tmp


# def find_file_positions(read_pos_list, mummer_out_path, mummer_out_path_rc):
#     print(read_pos_list)






import os
import sys

from collections import defaultdict
from time import time
from array import array
import dill as pickle 
import gffutils
import pysam


import signal
from multiprocessing import Pool
from operator import attrgetter

from modules import colinear_solver 
from modules import help_functions
from modules import classify_read_with_mams
from modules import classify_alignment2
from modules import sam_output
from modules import mem_wrapper


############### TMP #######
from types import ModuleType, FunctionType
from gc import get_referents

# Custom objects know their class.
# Function objects seem to know way too much, including modules.
# Exclude modules as well.
BLACKLIST = type, ModuleType, FunctionType


def getsize(obj):
    """sum size of object & members."""
    if isinstance(obj, BLACKLIST):
        raise TypeError('getsize() does not take argument of type: '+ str(type(obj)))
    seen_ids = set()
    size = 0
    objects = [obj]
    while objects:
        need_referents = []
        for obj in objects:
            if not isinstance(obj, BLACKLIST) and id(obj) not in seen_ids:
                seen_ids.add(id(obj))
                size += sys.getsizeof(obj)
                need_referents.append(obj)
        objects = get_referents(*need_referents)
    return size
######################

def import_data(args):

    if args.index:
        index_folder = args.index
    else:
        index_folder = args.outfolder
        
    ref_segment_sequences = help_functions.pickle_load( os.path.join(index_folder, 'ref_segment_sequences.pickle') )
    ref_exon_sequences = help_functions.pickle_load( os.path.join(index_folder, 'ref_exon_sequences.pickle') )
    ref_flank_sequences = help_functions.pickle_load( os.path.join(index_folder, 'ref_flank_sequences.pickle') )
    splices_to_transcripts = help_functions.pickle_load( os.path.join(index_folder, 'splices_to_transcripts.pickle') )
    transcripts_to_splices = help_functions.pickle_load( os.path.join(index_folder, 'transcripts_to_splices.pickle') )
    all_splice_pairs_annotations = help_functions.pickle_load( os.path.join(index_folder, 'all_splice_pairs_annotations.pickle') )
    all_splice_sites_annotations = help_functions.pickle_load( os.path.join(index_folder, 'all_splice_sites_annotations.pickle') )
    parts_to_segments = help_functions.pickle_load( os.path.join(index_folder, 'parts_to_segments.pickle') )
    segment_to_gene = help_functions.pickle_load( os.path.join(index_folder, 'segment_to_gene.pickle') )
    gene_to_small_segments = help_functions.pickle_load( os.path.join(index_folder, 'gene_to_small_segments.pickle') )
    max_intron_chr = help_functions.pickle_load( os.path.join(index_folder, 'max_intron_chr.pickle') )

    chr_to_id = help_functions.pickle_load( os.path.join(index_folder, 'chr_to_id.pickle') )
    id_to_chr = help_functions.pickle_load( os.path.join(index_folder, 'id_to_chr.pickle') )

    # tiling_ref_segment_sequences = help_functions.pickle_load( os.path.join(args.outfolder, 'tiling_ref_segment_sequences.pickle') )
    # tiling_parts_to_segments = help_functions.pickle_load( os.path.join(args.outfolder, 'tiling_parts_to_segments.pickle') )
    # tiling_segment_to_gene = help_functions.pickle_load( os.path.join(args.outfolder, 'tiling_segment_to_gene.pickle') )
    # tiling_gene_to_small_segments = help_functions.pickle_load( os.path.join(args.outfolder, 'tiling_gene_to_small_segments.pickle') )
    # tiling_structures = [tiling_segment_to_gene, tiling_parts_to_segments, tiling_gene_to_small_segments, tiling_ref_segment_sequences]

    # print("ref_segment_sequences:", getsize(ref_segment_sequences)//1000000)
    # print("ref_exon_sequences:", getsize(ref_exon_sequences)//1000000)
    # print("ref_flank_sequences:", getsize(ref_flank_sequences)//1000000)
    # print("splices_to_transcripts:", getsize(splices_to_transcripts)//1000000)
    # print("transcripts_to_splices:", getsize(transcripts_to_splices)//1000000)
    # print("all_splice_pairs_annotations:", getsize(all_splice_pairs_annotations)//1000000)
    # print("all_splice_sites_annotations:", getsize(all_splice_sites_annotations)//1000000)
    # print("parts_to_segments:", getsize(parts_to_segments)//1000000)
    # print("segment_to_gene:", getsize(segment_to_gene)//1000000)
    # print("gene_to_small_segments:", getsize(gene_to_small_segments)//1000000)
    # print("max_intron_chr:", getsize(max_intron_chr)//1000000)
    # print("chr_to_id:", getsize(chr_to_id)//1000000)
    # print("id_to_chr:", getsize(id_to_chr)//1000000)

    # print("tiling_ref_segment_sequences:", getsize(tiling_ref_segment_sequences)//1000000)
    # print("tiling_parts_to_segments:", getsize(tiling_parts_to_segments)//1000000)
    # print("tiling_segment_to_gene:", getsize(tiling_segment_to_gene)//1000000)
    # print("tiling_gene_to_small_segments:", getsize(tiling_gene_to_small_segments)//1000000)

    # return ref_segment_sequences, ref_flank_sequences, splices_to_transcripts, \
    #         transcripts_to_splices, all_splice_pairs_annotations, \
    #         all_splice_sites_annotations, parts_to_segments,\
    #         segment_to_gene, gene_to_small_segments, max_intron_chr, \
    #         ref_exon_sequences, chr_to_id, id_to_chr, tiling_structures

    return ref_segment_sequences, ref_flank_sequences, splices_to_transcripts, \
            transcripts_to_splices, all_splice_pairs_annotations, \
            all_splice_sites_annotations, parts_to_segments,\
            segment_to_gene, gene_to_small_segments, max_intron_chr, \
            ref_exon_sequences, chr_to_id, id_to_chr

def annotate_guaranteed_optimal_bound(mems, is_rc, max_intron_chr, max_global_intron):
    """
        Calculate the maximum coverage (mem-score) that a read can get per chromosome
        and annotate this value to each instance. We can use this annotation to avoid expensive MAM calculation 
        because we can basically continue to next read if the theoretically possibel maximum chaining value 
        for a chromosome is smaller than a solution already computed
    """
    # print("max_intron_chr", max_intron_chr)
    # split mems into separate instances if there is more than max_allowed_intron nt between two consecutive hits!
    # need to reindex j based on splitting the solution per chromosome..
    all_mem_sols = {} 
    for chr_id, all_mems_to_chromosome in mems.items():
        max_allowed_intron = min(max_intron_chr[chr_id] + 20000, max_global_intron)
        # print("max_allowed_intron",chr_id, max_allowed_intron)
        chr_instance_index = 0
        j_reindex = 0
        if len(all_mems_to_chromosome) > 1:
            m1 = all_mems_to_chromosome[0]
            m1 = m1._replace(j = j_reindex)
            # m1.j = j_reindex 
            curr_instance = [m1]
            for m1,m2 in zip(all_mems_to_chromosome[:-1], all_mems_to_chromosome[1:]):
                if m2.x - m1.y > max_allowed_intron:
                    all_mem_sols[(chr_id, chr_instance_index)] = curr_instance
                    chr_instance_index  += 1 
                    j_reindex = 0
                    m2 = m2._replace(j = j_reindex)
                    curr_instance = [m2]

                else:
                    j_reindex += 1 
                    # print(m2)
                    m2 = m2._replace(j = j_reindex)
                    # print(m2)
                    # print()
                    curr_instance.append(m2)
            all_mem_sols[(chr_id, chr_instance_index)] = curr_instance
                       
        else:
            all_mem_sols[(chr_id, chr_instance_index)] = all_mems_to_chromosome


    # Annotate maximum possible solution to each instance
    upper_bound = {}

    for (chr_id, chr_instance_index), all_mems_to_solution in all_mem_sols.items():
        starts = [("start", m.c) for m in all_mems_to_solution]
        stops = [("stop", m.d) for m in all_mems_to_solution]
        all_starts_stops = sorted(list(starts) + list(stops), key = lambda x: x[1])
        assert all_starts_stops[0][0] == 'start'
        active_start = all_starts_stops[0][1]
        nr_actives = 1
        intervals = []
        for site, pos in all_starts_stops[1:]:
            if nr_actives == 0:
                active_start = pos

            if site == 'stop':
                nr_actives -= 1
            else:
                nr_actives += 1

            if nr_actives == 0:
                assert site == 'stop'
                intervals.append( (active_start, pos) )

            assert nr_actives >= 0

        tot_cov = 0
        for start, stop in intervals:
            tot_cov += stop - start  + 1 # MEM choordinates are inclusive

        upper_bound[(chr_id, chr_instance_index)] = (tot_cov, is_rc, all_mems_to_solution)
    # for i in upper_bound:
    #     print("sols:", i, upper_bound[i][0], len(upper_bound[i][2]))

    return upper_bound


def find_exons(chr_id, mam_solution, ref_exon_sequences, ref_segment_sequences, \
                 ref_flank_sequences, all_splice_pairs_annotations):
    # print("BEFO:", [(m.x,m.y) for m in mam_solution])
    valid_introns_sites = all_splice_pairs_annotations[chr_id]
    # print("valid_introns_sites", valid_introns_sites)
    # identify known intron sites
    parts = []
    prev_split_index = 0
    if len(mam_solution) > 1:
        for i, (m1,m2) in enumerate(zip(mam_solution[:-1], mam_solution[1:])):
            if (m1.y, m2.x) in valid_introns_sites:
                parts.append( mam_solution[prev_split_index:i+1] )
                prev_split_index = i+1
        if mam_solution[prev_split_index:]:
            parts.append(mam_solution[prev_split_index:])
        # for p in parts:
        #     print(p)
    else:
        parts.append(mam_solution)
    
    # print("parts",parts)
    # print()
    # identify likely intron sites too large to be a deletion
    parts2 = []
    for part in parts:
        prev_split_index = 0
        if len(part) > 1:
            for i, (m1,m2) in enumerate(zip(part[:-1], part[1:])):
                if m2.x - m1.y > 10:
                    parts2.append( part[prev_split_index:i+1] )
                    prev_split_index = i+1
            if part[prev_split_index:]:
                # print("GAWWWWD", part[prev_split_index:])
                parts2.append(part[prev_split_index:])
        else:
            parts2.append(part)

    # for p in parts2:
    #     print("parts", p)

    # finally look for exact exons or enclosing exons if found
    exons = []
    for part in parts2:
        # check if all segments are adjacent 
        is_adjacent = False
        if len(part) == 1:
            is_adjacent = True
        else: 
            is_adjacent = all([ m1.y == m2.x for (m1,m2) in zip(part[:-1], part[1:]) ])

        if is_adjacent:
            for j, mam in enumerate(part):
                exons.append((mam.x, mam.y, mam.c, mam.d, mam.ref_chr_id))
        else:
            # check if there is an combination of segments and/or adjacent exons over the total region spanned by the exons 

            all_points = []
            segm = {}
            start_points = {}
            end_points = {}
            for m in part:
                all_points.append(m.x)
                all_points.append(m.y)
                segm[(chr_id, m.x, m.y)] = m
                start_points[m.x] = m
                end_points[m.y] = m
            cover = {}
            for i, p1 in enumerate(sorted(all_points)):
                cover[p1] = []
                for j, p2 in enumerate(sorted(all_points)):
                    key = array('L', [chr_id, p1, p2])
                    if key.tobytes() in ref_exon_sequences or  (chr_id, p1, p2) in segm:
                        cover[p1].append(p2)


            # print("all_points", all_points)
            # print("all covers", cover)
            paths = help_functions.find_all_paths(cover, part[0].x, part[-1].y)
            # print(start_points)
            # print(end_points)

            if len(paths) > 0:
                # print("paths[0]" , paths[0])
                path = paths[0]
                for (p1, p2) in zip(path[:-1], path[1:]):
                    if (chr_id, p1, p2) in segm:
                        mam = segm[(chr_id, p1, p2)]
                        exons.append((mam.x, mam.y, mam.c, mam.d, mam.ref_chr_id))
                    else:
                        if p1 in start_points:
                            m = start_points[p1]
                            c = m.c
                            x = m.x
                        else:
                            m = end_points[p1]
                            c = m.d
                            x = m.y 

                        if p2 in start_points:
                            m = start_points[p2]
                            d = m.c
                            y = m.x
                        else:
                            m = end_points[p2]
                            d = m.d
                            y = m.y                            
                        exons.append((x, y, c, d, chr_id))

            else:  #if not, simply give up and put the small spurious intron caused by optimal solution is not containing adjacent segments
                for j, mam in enumerate(part):
                    exons.append((mam.x, mam.y, mam.c, mam.d, mam.ref_chr_id))  

    # print("LOOL NEW",[ (x,y) for x, y, c, d, seq_id  in  exons])

    chained_exon_seqs = []
    predicted_exons = []
    prev_y_coord = -1
    covered = 0
    # for mam in mam_solution:
    for x, y, c, d, seq_id in exons:
        key_tmp = array('L', [seq_id, x, y])
        key = key_tmp.tobytes()
        if key in ref_exon_sequences:
            seq = ref_exon_sequences[key] 
            covered += d - c + 1
        else: 
            if key in ref_segment_sequences:
                seq = ref_segment_sequences[key] 
                covered += d - c + 1
            elif key in ref_flank_sequences:
                seq = ref_flank_sequences[key] 
                covered += d - c + 1
            else:
                print("Bug encountered, {0} is not in {1}".format((x, y), mam_solution))

        if prev_y_coord == x: #adjacent segments means its a flank and we should not add an new exon (i.e., intron split)
            predicted_exons[-1] = (predicted_exons[-1][0], y)  # update the last exon
        else:
            predicted_exons.append( (x, y) )

        prev_y_coord = y
        chained_exon_seqs.append(seq)
    created_ref_seq = "".join([exon for exon in chained_exon_seqs])
    predicted_splices = [ (e1[1],e2[0]) for e1, e2 in zip(predicted_exons[:-1],predicted_exons[1:])]
    return exons, created_ref_seq, predicted_exons, predicted_splices, covered


def get_exact_alignment(read_seq, created_ref_seq, mam_sol_exons_length):
    # if statement: faster and more memory efficient alignments with edlib if sequences are very long. 
    # However, much worse quality.
    if len(created_ref_seq) > 20000 or len(read_seq) > 20000 or (1000 < len(read_seq) < mam_sol_exons_length/10): 
        # print("lenght ref: {0}, length query:{1}".format(len(created_ref_seq), len(read_seq)))
        read_aln, ref_aln, edit_distance = help_functions.edlib_alignment(read_seq, created_ref_seq, aln_mode = "HW")
        match_score = sum([2 for n1,n2 in zip(read_aln, ref_aln) if n1 == n2 ])
        # diff_score = sum([2 for n1,n2 in zip(read_aln, ref_aln) if n1 != n2 ])
        alignment_score = match_score - 2*edit_distance
        # print(read_seq)
    else:
        read_aln, ref_aln, cigar_string, cigar_tuples, alignment_score = help_functions.parasail_alignment(read_seq, created_ref_seq)
        # print(read_acc)
        # print(read_aln)
        # print(ref_aln)

    return read_aln, ref_aln, alignment_score 


# def run_tiling_solution(mem_solution, tiling_ref_segment_sequences, ref_flank_sequences, tiling_parts_to_segments, \
#                         tiling_segment_to_gene, tiling_gene_to_small_segments, \
#                         read_seq, warning_log_file, min_acc, \
#                         chr_id, ref_exon_sequences, \
#                         all_splice_pairs_annotations,
#                         splices_to_transcripts, transcripts_to_splices, \
#                         all_splice_sites_annotations, mam_sol_exons_length, \
#                         alignment_score, read_aln, ref_aln, current_classification, non_covered_regions, covered,
#                         predicted_exons, annotated_to_transcript_id):

#     non_covered_regions_tiling, mam_value_tiling, mam_solution_tiling = classify_read_with_mams.main(mem_solution, tiling_ref_segment_sequences, ref_flank_sequences, tiling_parts_to_segments, \
#                                                                                                             tiling_segment_to_gene, tiling_gene_to_small_segments, \
#                                                                                                             read_seq, warning_log_file, min_acc, is_tiling_instance = True)

#     # print("TILING finished Mam solution Tiling!!:",mam_value_tiling, mam_solution_tiling)
#     # for zzz2 in mam_solution_tiling:
#     #     print(zzz2)
#     exons, tiling_created_ref_seq, tiling_predicted_exons, tiling_predicted_splices, tiling_covered = find_exons(chr_id, mam_solution_tiling, ref_exon_sequences, \
#                                                                             tiling_ref_segment_sequences, ref_flank_sequences, all_splice_pairs_annotations)

#     tiling_classification, tiling_annotated_to_transcript_id = classify_alignment2.main(chr_id, tiling_predicted_splices, splices_to_transcripts, transcripts_to_splices, all_splice_pairs_annotations, all_splice_sites_annotations)

#     tiling_read_aln, tiling_ref_aln, tiling_alignment_score = get_exact_alignment(read_seq, tiling_created_ref_seq, mam_sol_exons_length)

#     if tiling_alignment_score > alignment_score or tiling_classification == 'FSM':
#         # print("TILING")        
#         # print("old aln:", read_aln)
#         # print("old aln:", ref_aln)
#         # print(tiling_classification)            
#         # print("BEFORE", non_covered_regions)
#         # print("new aln:", tiling_read_aln)
#         # print("new aln:", tiling_ref_aln)
#         # print("NOW", non_covered_regions_tiling) 
#         # print("read was", current_classification, "had aln score", alignment_score)
#         # print("tiling read is ", tiling_classification, "has aln score", tiling_alignment_score)
#         return  tiling_classification, tiling_alignment_score, non_covered_regions_tiling, \
#                 tiling_read_aln, tiling_ref_aln, tiling_predicted_exons, tiling_annotated_to_transcript_id, tiling_covered
#     else:
#         return current_classification, alignment_score, non_covered_regions, \
#                 read_aln, ref_aln, predicted_exons, annotated_to_transcript_id, covered


def align_single(reads, refs_lengths, args,  batch_number):
    # print("reads:", getsize(reads)//1000000)
    auxillary_data = import_data(args)
    # import time
    # time.sleep(10000)
    mems_path =  os.path.join( args.outfolder, "seeds_batch_{0}.txt".format(batch_number) )
    mems_path_rc =  os.path.join( args.outfolder, "seeds_batch_{0}_rc.txt".format(batch_number) )
    nlog_n_instance_counter = 0
    quadratic_instance_counter = 0
    max_global_intron = args.max_intron
    min_acc = args.min_acc
    ref_segment_sequences, ref_flank_sequences, splices_to_transcripts, \
    transcripts_to_splices, all_splice_pairs_annotations, \
    all_splice_sites_annotations, parts_to_segments, \
    segment_to_gene, gene_to_small_segments, max_intron_chr, \
    ref_exon_sequences, chr_to_id, id_to_chr = auxillary_data
    # ref_exon_sequences, chr_to_id, id_to_chr, tiling_structures = auxillary_data

    if batch_number == -1:
        alignment_outfile = pysam.AlignmentFile( os.path.join(args.outfolder, args.prefix+".sam"), "w", reference_names=list([id_to_chr[id_] for id_ in refs_lengths.keys()]), reference_lengths=list(refs_lengths.values()) ) #, template=samfile)
        warning_log_file = open(os.path.join(args.outfolder, "uLTRA.stderr"), "w")

    else:  
        alignment_outfile = pysam.AlignmentFile( os.path.join(args.outfolder, "reads_batch_{0}.sam".format(batch_number)), "w", reference_names=list([id_to_chr[id_] for id_ in refs_lengths.keys()]), reference_lengths=list(refs_lengths.values()) ) #, template=samfile)
        warning_log_file = open(os.path.join(args.outfolder, "uLTRA_batch_{0}.stderr".format(batch_number)), "w")



    # tiling_segment_to_gene, \
    # tiling_parts_to_segments, tiling_gene_to_small_segments, \
    # tiling_ref_segment_sequences = tiling_structures # unpacking tiling structures

    classifications = defaultdict(str)
    read_accessions_with_mappings = set()
    processed_read_counter = 0

    for (read_acc, mems), (_, mems_rc) in zip(mem_wrapper.get_mem_records(mems_path,reads), mem_wrapper.get_mem_records(mems_path_rc, reads)):
        # multiple = False
        if read_acc not in reads: # if parallelization not all reads in mummer file are in read batches
            continue
        else:
            read_seq_mod = help_functions.remove_read_polyA_ends(reads[read_acc], args.reduce_read_ployA, 1)
        # print("instance sizes fw:", [ (chr_id, len(mm)) for chr_id, mm in mems.items()])
        # print("instance sizes rc:", [ (chr_id, len(mm)) for chr_id, mm in mems_rc.items()])
        # print()
        # print()
        # print(read_acc)
        upper_bound = annotate_guaranteed_optimal_bound(mems, False, max_intron_chr, max_global_intron)
        upper_bound_rc = annotate_guaranteed_optimal_bound(mems_rc, True, max_intron_chr, max_global_intron)
        # print()
        processed_read_counter += 1
        if processed_read_counter % 5000 == 0:
            print('Processed {0} reads in batch {1}'.format(processed_read_counter, batch_number))
        # do the chaining here immediately!
        all_chainings = []
        best_solution_value = 0
        for (chr_id, chr_instance_index) , (upper_bound_cov, is_rc, all_mems_to_chromosome) in sorted(list(upper_bound.items()) + list(upper_bound_rc.items()), key = lambda x: x[1][0], reverse = True ): # mems.items():
            # print(read_acc, id_to_chr[chr_id], upper_bound_cov)
            if upper_bound_cov < best_solution_value*args.dropoff:
                # print("Breaking for", chr_id, is_rc, upper_bound_cov, "best:", best_solution_value, "read length:", len(read_seq_mod))
                break
            
            # print("Processing", chr_id, is_rc, upper_bound_cov, "best:", best_solution_value, "read length:", len(read_seq_mod))
            # for mem in all_mems_to_chromosome:
            #     print(mem.exon_part_id, mem.x, mem.y, mem.c, mem.d, '\t', mem.val)
            # print(len(all_mems_to_chromosome))


            max_allowed_intron = min(max_intron_chr[chr_id] + 20000, max_global_intron)
            # print("max_allowed_intron", max_allowed_intron, max_intron_chr[chr_id])
            # print("LEM MEMS", len(all_mems_to_chromosome))
            if len(all_mems_to_chromosome) < 90:
                solutions, mem_solution_value = colinear_solver.read_coverage(all_mems_to_chromosome, max_allowed_intron)
                quadratic_instance_counter += 1 
            else:
                solutions, mem_solution_value = colinear_solver.n_logn_read_coverage(all_mems_to_chromosome)
                nlog_n_instance_counter += 1

            if mem_solution_value > best_solution_value:
                best_solution_value = mem_solution_value
                # print("best now:", mem_solution_value)

            # if len(solutions) > 1:
            #     # print("More than 1 solution on chromosome")
            #     # for sol in solutions:
            #     #     print(mem_solution_value, [ (m.x, m.y) for m in sol])
            #     # multiple = True

            for sol in solutions:
                all_chainings.append( (chr_id, sol, mem_solution_value, is_rc) )
                # for mem in sol:
                #     print(mem.exon_part_id, mem.x, mem.y, mem.c, mem.d, '\t', mem.val, "(sol)")
            # print(all_chainings)
        is_secondary =  False
        is_rc =  False
        if not all_chainings:
            sam_output.main(read_acc, read_seq_mod, '*', 'unaligned', [], '*', '*', '*', alignment_outfile, is_rc, is_secondary, 0)
            continue

        # all_chainings = sorted(all_chainings, key=lambda x: x[2], reverse=True) 
        all_chainings = sorted(all_chainings, key=lambda x:(-x[2],(x[1][-1].y - x[1][0].x))) 
        # if multiple:
        #     print(all_chainings)
        best_chaining_score = all_chainings[0][2]
        read_alignments = []
        mam_solutions = set()
        for i_nr_sol, (chr_id, mem_solution, chaining_score, is_rc) in enumerate(all_chainings):
            # print(i_nr_sol, chr_id, chaining_score)
            # if read_acc == "100:823|c8740d7a-53bd-4690-aa53-de3ebd003d20": #len(all_chainings) > 20: 
            #     print(read_acc, len(all_chainings), chaining_score)
            #     print(all_chainings)
            # for c in all_chainings:
            #     print(c[1][-1].y - c[1][0].x, c[1][0].x, c[1][-1].y)
            #     sys.exit()
            if chaining_score/float(best_chaining_score) < args.dropoff or i_nr_sol >= args.max_loc:
                # print(chr_id, chaining_score, best_chaining_score, "NOT CONSIDERED")
                # print(i_nr_sol, "lol")
                break
            # print(chr_id, chaining_score, best_chaining_score)

            if is_rc:
                read_seq = help_functions.reverse_complement(read_seq_mod)
            else:
                read_seq = read_seq_mod
            # print("mem solution:", is_rc, chaining_score, mem_solution)
            # print()
            # print()
            # print(is_rc, mem_solution[0].x)
            # print(mem_solution)
            # print(read_seq)
            non_covered_regions, mam_value, mam_solution = classify_read_with_mams.main(mem_solution, ref_segment_sequences, ref_flank_sequences, parts_to_segments, \
                                                                                                                    segment_to_gene, gene_to_small_segments, \
                                                                                                                    read_seq, warning_log_file, min_acc)
            
            # We can enter the if statement below because sometimes the MEM chaining finder will 
            # return multiple optimal chainings that lead to the same mam_solution
            if mam_solution in mam_solutions:
                continue

            mam_solutions.add(mam_solution)
            # print("finished Mam solution:",mam_value, mam_solution)
            # for zzz2 in mam_solution:
            #     print(zzz2)
            # print(non_covered_regions)
            mam_sol_exons_length = sum([ mam.y - mam.x for mam in mam_solution])
            # print(max_intron_size)
            if mam_value > 0:
                exons, created_ref_seq, predicted_exons, predicted_splices, covered = find_exons(chr_id, mam_solution, ref_exon_sequences, \
                                                                                            ref_segment_sequences, ref_flank_sequences, all_splice_pairs_annotations)


                classification, annotated_to_transcript_id = classify_alignment2.main(chr_id, predicted_splices, splices_to_transcripts, transcripts_to_splices, all_splice_pairs_annotations, all_splice_sites_annotations)
                largest_intron_size = max([m2.x - m1.y for m1,m2 in zip(mam_solution[:-1], mam_solution[1:]) ]) if len(mam_solution) > 1 else 0
                if largest_intron_size > max_allowed_intron and classification != 'FSM':
                    # print(i_nr_sol, mem_solution[0].x, "HERE", largest_intron_size)
                    # print()
                    # print(read_acc)
                    # print(classification, alignment_score/len(read_seq), "Score: {0}, old T: {1}, new T: {2}".format(alignment_score, 2*args.alignment_threshold*len(read_seq), len(read_seq)*8*args.alignment_threshold))
                    continue

                read_aln, ref_aln, alignment_score = get_exact_alignment(read_seq, created_ref_seq, mam_sol_exons_length)

                # if classification != 'FSM' and len(non_covered_regions) >= 3 and (max(non_covered_regions[1:-1]) > args.non_covered_cutoff):
                #     classification, alignment_score, non_covered_regions, \
                #     read_aln, ref_aln, predicted_exons, annotated_to_transcript_id, covered = run_tiling_solution(mem_solution, tiling_ref_segment_sequences, ref_flank_sequences, tiling_parts_to_segments, \
                #                         tiling_segment_to_gene, tiling_gene_to_small_segments, \
                #                         read_seq, warning_log_file, min_acc, chr_id, ref_exon_sequences, \
                #                         all_splice_pairs_annotations,
                #                         splices_to_transcripts, transcripts_to_splices, \
                #                         all_splice_sites_annotations, mam_sol_exons_length, alignment_score, \
                #                         read_aln, ref_aln, classification, non_covered_regions, covered, predicted_exons, annotated_to_transcript_id)



                if alignment_score < 2*args.alignment_threshold*len(read_seq) and classification != 'FSM': # match score * aln_threshold
                    # print(i_nr_sol, mem_solution[0].x, "HERE2", chaining_score, alignment_score, 2*args.alignment_threshold*len(read_seq))
                    # print()
                    # print(read_acc)
                    # print(read_aln)
                    # print(ref_aln)
                    # print(classification, alignment_score/len(read_seq), "Score: {0}, T: {1}".format(alignment_score, 2*args.alignment_threshold*len(read_seq)))
                    continue
                # print(classification)
                # checing for internal (splice site) non covered regions
                if len(non_covered_regions) >= 3 and (max(non_covered_regions[1:-1]) > args.non_covered_cutoff):
                    classification = 'Insufficient_junction_coverage_unclassified'
                coverage = covered / float(len(read_seq)) 
                # print(len(read_alignments), read_alignments)
                genome_start = mam_solution[0].x
                genome_stop = mam_solution[-1].y
                read_alignments.append( (alignment_score, genome_start, genome_stop, read_acc, chr_id, classification, predicted_exons, read_aln, ref_aln, annotated_to_transcript_id, is_rc, coverage) )
            
            # else:
            #     print(i_nr_sol, mem_solution[0].x, "OK", chaining_score, mam_value, mem_solution[-1].y - mem_solution[0].x, mem_solution[0].x)    
        # if read_acc == "100:823|c8740d7a-53bd-4690-aa53-de3ebd003d20": #len(all_chainings) > 20: 
        #     print(read_acc, len(all_chainings), chaining_score, chr_id, max_allowed_intron)
        #     print(read_seq)
        #     # print(all_chainings)
        #     for c in all_chainings:
        #         print(is_rc, c[2], c[1][-1].y - c[1][0].x, c[1][0].x, c[1][-1].y)
        #     print("Nr alignments:", len(read_alignments))
        #     for r in read_alignments:
        #         print(r[0],r[1], r[2], r[3],r[4])
        #     # sys.exit()

            # elif 10*len(read_seq) < mam_sol_exons_length:
            #     print("length ref: {0}, length query:{1}".format(mam_sol_exons_length, len(read_seq)))
            #     print(read_acc, "to chr", chr_id)
            #     print(read_seq)

        ##################  Process alignments and decide primary
        if len(read_alignments) == 0:
            sam_output.main(read_acc, read_seq, '*', 'unaligned', [], '*', '*', '*', alignment_outfile, is_rc, is_secondary, 0)
        else:
            # sorted_wrt_alignement_score = sorted(read_alignments, key = lambda x: x[0], reverse = True)
            sorted_wrt_alignement_score = sorted(read_alignments, key = lambda x: (-x[0], (x[2] - x[1]), x[5]))
            # sorted_wrt_alignement_score = sorted(read_alignments, key = lambda x: (-(x[0] - 0.002*(x[2] - x[1])), len(x[5]) ))
            # if len(read_alignments) > 1 and "FSM" in set( [r[5] for r in read_alignments]):
            #     print(read_acc)
            #     for r in sorted_wrt_alignement_score:
            #         print(r[0], r[5], r[1], (r[2] - r[1]))
            #         print(r[7])
            #         print(r[8])
            #     for c in all_chainings:
            #         print(c[0], c[2], c[1][0].x, c[1][-1].y)
            # sorted_wrt_alignement_score = sorted(read_alignments, key = lambda x: (x[5] != "FSM", -x[0], (x[2] - x[1])))
            best_aln_sw_score = sorted_wrt_alignement_score[0][0]
            more_than_one_alignment = True if len(sorted_wrt_alignement_score) > 1 else False

            # if read_acc == "100:823|c8740d7a-53bd-4690-aa53-de3ebd003d20": #len(all_chainings) > 20: 
            #     print()
            #     print( "SORTED", best_aln_sw_score)
            #     for r in sorted_wrt_alignement_score:
            #         print(r[0],r[1], (r[2]-r[1]), r[5],r[6])
            #     sys.exit()


            for i, (alignment_score, genome_start, genome_stop, read_acc, chr_id, classification, predicted_exons, read_aln, ref_aln, annotated_to_transcript_id, is_rc, coverage) in enumerate(sorted_wrt_alignement_score):
                if i == 0:
                    is_secondary =  False
                    assert alignment_score == best_aln_sw_score
                    if more_than_one_alignment:
                        if alignment_score == sorted_wrt_alignement_score[1][0]:
                            map_score = 0
                        else:
                            map_score = 60                           
                    else:
                        map_score = 60
                    classifications[read_acc] = (classification, coverage )
                else:
                    is_secondary =  True
                    map_score = 0


                sam_output.main(read_acc, read_seq, id_to_chr[chr_id], classification, predicted_exons, read_aln, ref_aln, annotated_to_transcript_id, alignment_outfile, is_rc, is_secondary, map_score, aln_score = alignment_score)
                read_accessions_with_mappings.add(read_acc)


    alignment_outfile.close()
    warning_log_file.close()
    print("Number of instances solved with quadratic collinear chainer solution:", quadratic_instance_counter)
    print("Number of instances solved with n*log n collinear chainer solution:", nlog_n_instance_counter)
    # print(alignment_outfile.filename, dir(alignment_outfile))
    return classifications, alignment_outfile.filename




def align_single_helper(arguments):
    # for v in arguments:
    #     args = v
    # print(len(arguments))
    return align_single(*arguments)
    # print(len(args))
    # read_data, auxillary_data = args[0], args[1]
    # align_single()
    # return 1



def align_parallel(read_data, refs_lengths, args):
    ####### parallelize alignment #########
    # pool = Pool(processes=mp.cpu_count())
    original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGINT, original_sigint_handler)

    start_multi = time()
    pool = Pool(processes=int(args.nr_cores))
    try:
        res = pool.map_async(align_single_helper, [ (d, refs_lengths, args, i) for i,d in enumerate(read_data)] )
        results =res.get(999999999) # Without the timeout this blocking call ignores all signals.
    except KeyboardInterrupt:
        print("Caught KeyboardInterrupt, terminating workers")
        pool.terminate()
        sys.exit()
    else:
        pool.close()
    pool.join()

    classifications = defaultdict(str)
    alignment_outfiles = []
    for r, alignment_outfile_name in results:
        alignment_outfiles.append(alignment_outfile_name)
        for acc in r:
            classifications[acc] = r[acc]
        # print(r)
    print("Time elapesd multiprocessing:", time() - start_multi)  
    return classifications, alignment_outfiles






from time import time
import queue 
import random
import copy




class Node:
    __slots__ = ['d', 'j', 'Cj', "j_max"]
    def __init__(self, d, j, Cj, j_max):
        self.d = d
        self.j = j
        self.Cj = Cj
        self.j_max = j_max


def construct_tree(tree, leafs, n):   
    # assign values to leaves of  
    # the segment tree  
    for i in range(n):  
        tree[n + i] = leafs[i] 
      
    # assign values to remaining nodes  
    # to compute maximum in a given range  
    for i in range(n - 1, 0, -1):  
        # print(i)
        # print(tree)
        max_coord_node = max([tree[2 * i ], tree[2 * i + 1]], key = lambda x: x.d ) # should always be left one though
        tree_node = Node(max_coord_node.d, max_coord_node.j, max_coord_node.Cj, max_coord_node.j_max)
        tree[i] = tree_node


def range_query(tree, l, r, n):
    assert l <= r
    # search for both at the same time 
    l_pos = 1 # root position
    r_pos = 1 # root position
    V_biss = set()
    V_prime = set()
    v = 1
    while True: # not yet reached a leaf

        l_pos_l_child = 2*l_pos
        l_pos_r_child = 2*l_pos + 1
        
        r_pos_l_child = 2*r_pos
        r_pos_r_child = 2*r_pos + 1

        ## traversing downwards for right path
        if r >= tree[r_pos_l_child].d:
            r_pos = r_pos_r_child
        else: # tree[left_child_index].d <= c:
            r_pos = r_pos_l_child         

        ## traversing downwards for left path
        if  l > tree[l_pos_l_child].d:
            l_pos = l_pos_r_child 
        else:
            l_pos = l_pos_l_child 

        if l_pos == r_pos:
            v = l_pos

        ## Adding nodes to sets V' and V'' if the two paths separates into two (See page 20-21 in GSAD book, Makinen et al)
        if l_pos != r_pos and v != l_pos >> 1: # paths have split and v is not the immediate predecendant
            if r_pos == r_pos_r_child: # we took a right step for right path
                V_biss.add(r_pos_l_child)

            if l_pos == l_pos_l_child: # we took a left step for left path
                V_prime.add(l_pos_r_child)
        
        # exit if we are at leafs
        if l_pos >= n and r_pos >= n:
            break

    R = tree[r_pos].d
    if R == r:
        V_biss.add(r_pos)
    L = tree[l_pos].d
    if L >= l:
        V_prime.add(l_pos)

    # #####
    # tmp = sorted(V_prime, key = lambda x: tree[x].Cj)
    # if len(tmp) > 1:
    #     if tree[ tmp[-1] ].Cj  == tree[ tmp[-2] ].Cj:
    #         chosen = max(V_prime, key = lambda x: tree[x].Cj)
    #         all_ = [zz for zz in V_prime if tree[ chosen ].Cj == tree[ zz ].Cj]
    #         print("LOOOOOL", tree[ chosen ].j_max, [tree[ zz ].j_max for zz in all_])
    
    # tmp = sorted(V_biss, key = lambda x: tree[x].Cj)
    # if len(tmp) > 1:
    #     if tree[ tmp[-1] ].Cj  == tree[ tmp[-2] ].Cj:
    #         chosen = max(V_biss, key = lambda x: tree[x].Cj)
    #         all_ = [zz for zz in V_biss if tree[ chosen ].Cj == tree[ zz ].Cj]
    #         print("LOOOOOL", tree[ chosen ].j_max, [tree[ zz ].j_max for zz in all_])
    # ########


    if V_prime:
        # vl = max(V_prime, key = lambda x: tree[x].Cj)
        # The sorted function guarantees largest j_max is returned to be consistent (i.e., closest mem)
        vl = max(sorted(V_prime, key = lambda x: tree[x].j_max, reverse = True), key = lambda x: tree[x].Cj)
    else:
        vl = None

    if V_biss:
        # vr = max(V_biss, key = lambda x: tree[x].Cj)
        # The sorted function guarantees largest j_max is returned to be consistent (i.e., closest mem)
        vr = max(sorted(V_biss, key = lambda x: tree[x].j_max, reverse = True), key = lambda x: tree[x].Cj)

    else:
        vr = None
       
    if vl is not None and vr is not None:
        v_max_pos = max(sorted([vl,vr], key = lambda x: tree[x].j_max, reverse = True), key = lambda x: tree[x].Cj)
        # v_max_pos = max([vl,vr], key = lambda x: tree[x].Cj)
        C_max = tree[v_max_pos].Cj
    elif vl is not None:
        v_max_pos = vl
        C_max = tree[vl].Cj
    elif vr is not None:
        v_max_pos = vr
        C_max = tree[vr].Cj
    else:
        print("BUG", l,r)
        print(tree)
        sys.exit()
    
    # print(l,r , "v", v, "RQ2",V_prime, V_biss, "RQ2",C_max)
    return C_max, tree[v_max_pos].j_max, v_max_pos



def update(tree, leaf_pos, value, n): 
    # change the index to leaf node first  
    pos = leaf_pos + n
    # update the value at the leaf node  
    # at the exact index 
    tree[pos].Cj = value  
    while (pos > 1): 
          
        # move up one level at a time in the tree  
        pos >>= 1  
        # print('updating: ', pos)
        # update the values in the nodes  
        # in the next higher level 
        # cur_best = max([tree[2 * pos], tree[2 * pos + 1]], key=lambda x: x.Cj)
        cur_best = max( sorted([tree[2 * pos], tree[2 * pos + 1]], key = lambda x: x.j_max, reverse = True), key=lambda x: x.Cj)
        # if tree[2 * pos].Cj == tree[2 * pos + 1].Cj:
        #     print("OMGGGG",cur_best.j_max, tree[2 * pos].j_max, tree[2 * pos + 1].j_max)
        # tree[pos].Cj = max(tree[2 * pos].Cj,  
        #                    tree[2 * pos + 1].Cj)  
        tree[pos].Cj = cur_best.Cj
        tree[pos].j_max = cur_best.j_max




def argmax(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])[0]

def max_both(iterable):
    return max(enumerate(iterable), key=lambda x: x[1])


def reconstruct_solution(mems, C, trace_vector):
    solution_index = argmax(C)
    value = C[solution_index]
    # print()
    solution = []
    while solution_index > 0:
        solution.append(mems[solution_index - 1])  # trace vector is shifted on to the right so need to remove 1 from vectore to get j_index 
        solution_index = trace_vector[solution_index]
        # print(solution_index)
        # if solution_index is None:
        #     break
    return value, solution[::-1]

def reconstruct_all_solutions(mems, all_C_max_indicies, trace_vector):
    # solution_index = argmax(C)
    solutions = []
    for solution_index in all_C_max_indicies:
        value = C[solution_index]
        solution = []
        while solution_index > 0:
            solution.append(mems[solution_index - 1])  # trace vector is shifted on to the right so need to remove 1 from vectore to get j_index 
            solution_index = trace_vector[solution_index]
        solutions.append( solution[::-1] )
    return value, solutions

def make_leafs_power_of_2(mems):
    nodes = []
    nodes.append( Node(-1, -1, -2**32, -1) ) # add an start node in case a search is smaller than any d coord in tree
    for i, mem in enumerate(mems):
        # if i > 3: break
        m = Node(mem.d, mem.j, -2**32, mem.j)
        nodes.append(m)

    for i in range(20):
        if len(nodes) == 2**i or len(nodes) == 2**(i+1):
            break
        elif 2**i < len(nodes) < 2**(i+1):
            remainder = 2**(i+1) - len(nodes) 
            # print(remainder,  2**(i+1), len(nodes))
            for j in range(remainder):
                nodes.append( Node(-j-1, -j - 2, -2**32, -j - 2) ) # fill up nodes to have leaves a power of 2
            break

    leafs = sorted(nodes, key= lambda x: x.d)
    # n = len(leafs)
    return leafs


def all_solutions_c_max_indicies(C, C_max):
    return [i for i, c in enumerate(C) if c == C_max] 

## driver code for test if called as external script 

if __name__ == '__main__':
    from collections import namedtuple
    import colinear_solver 

    # construct sorted leafs
    
    def generate_mems(nr_mems):
        mem = namedtuple('Mem', ['x', 'y', 'c', 'd', 'val','j'])
        mem_choords = []
        for i in range(nr_mems):
            ref_start = random.randint(1,10000000)
            mem_length = random.randint(10,10)
            ref_stop = ref_start + mem_length
            read_start = random.randint(1,10000000)
            m = (ref_start, ref_stop,  read_start, read_start + mem_length, mem_length)
            mem_choords.append(m)
        mem = namedtuple('Mem', ['x', 'y', 'c', 'd', 'val','j'])
        mems = []
        for i, m in enumerate(sorted(mem_choords, key=lambda x: x[1])):
            m = mem(m[0], m[1],  m[2], m[3], m[4], i)
            mems.append(m)
        return mems

    mems = generate_mems(100)
    
    mem = namedtuple('Mem', ['x', 'y', 'c', 'd', 'val','j'])
    mems = [ mem(1, 10,  1, 10, 10, 0), mem(11, 20,  11, 20, 10, 1),
            mem(21, 30,  1, 10, 10, 2), mem(31, 40,  11, 20, 10, 3)]

    
    st = time()
    solution_quadratic, mem_solution_value = colinear_solver.read_coverage(mems)
    for sol in solution_quadratic:
        print(mem_solution_value, [mem.j for mem in sol])
    print("Total time quadratic method:", time()- st)
    print()


    st = time()
    T_leafs = make_leafs_power_of_2(mems)
    I_leafs = make_leafs_power_of_2(mems)
    n = len(T_leafs)
    mem_to_leaf_index = {l.j : i for i,l in enumerate(T_leafs)}
    T = [0 for i in range(2 * n) ]  
    I = [0 for i in range(2 * n) ]  
    construct_tree(T, T_leafs, n)
    construct_tree(I, I_leafs, n)
    time_construct = time()- st
    print("Time constructing tree:", time_construct)


    ############  T and I ###################
    #########################################
    st = time()
    # leafs = sorted(copy.deepcopy(nodes), key= lambda x: x.d)
    # I_leafs = copy.deepcopy(leafs)
    # for leaf in I_leafs:
    #     leaf.Cj -= leaf.d 
    # print([l.Cj for l in I_leafs])

    C = [0]* (len(mems) + 1) #(len(leafs))
    trace_vector = [None]*(len(mems) + 1)
    # print("leaf index (first mem):",mem_to_leaf_index[0],mem_to_leaf_index[-1],mem_to_leaf_index[1], "total leafs",len(mem_to_leaf_index), "len mems:", len(mems))
    update(T, mem_to_leaf_index[-1], 0, n) # point update 
    update(I, mem_to_leaf_index[-1], 0, n) # point update 

    for j, mem in enumerate(mems):
        # print(mem)
        # print("vals T:", [l.Cj for l in leafs])
        # print("vals I:", [l.Cj for l in I_leafs])
        leaf_to_update = mem_to_leaf_index[j]
        # print()
        c = mem.c
        T_max, j_prime_a, node_pos  = range_query(T, -1, c-1, len(T_leafs)) 
        # print("C_a:",  T_max +  mem.d - mem.c + 1, j_prime_a, node_pos, leaf_to_update )
        # print("T TREE:", [(s, zz.j, zz.d, zz.Cj, zz.j_max) for s, zz in enumerate(T) if type(zz) != int])
        C_a =  T_max +  mem.d - mem.c + 1  # add the mem_length to T since disjoint

        if T_max < 0:
            print("BUG", T_max)
            sys.exit()

        
        d = mem.d
        I_max, j_prime_b, node_pos  = range_query(I, c, d, len(I_leafs))         
        # print("C_b:", I_max +  mem.d, I_max, j_prime_b, node_pos, leaf_to_update )
        # print( I_max, mem.d, mems[j_prime_b].d, mems[j_prime_b])
        # print("I TREE:", [(s, zz.j, zz.d, zz.Cj, zz.j_max) for s, zz in enumerate(I) if type(zz) != int])
        C_b =  I_max +  mem.d #- mems[j_prime_b].d   # add the part of the mem that is not overlapping

        # if C_b < 0:
        #     print("BUG")
        #     sys.exit()
        # if C_a == C_b:
        #     print(j, C_b, "LOOOOL solve this by taking the j that is largest!",j_prime_a, j_prime_b)
        #     # j_prime = max(j_prime_a, j_prime_b)

        # else:
        index, value = max_both([C_a, C_b])
        if index == 0: # Updating with C_a
            j_prime = j_prime_a
        else: # Updating with C_b
            j_prime = j_prime_b

        C[j+1] = value

        if j_prime < 0: # any of the additional leaf nodes (with negative index) we add to make number of leafs 2^n
            trace_vector[j+1] = 0
        elif value == 0: # first j (i.e. j=0) 
            trace_vector[j+1]= 0
        else:
            trace_vector[j+1] = j_prime +1

        update(T, leaf_to_update, value, n) # point update 
        update(I, leaf_to_update, value - mem.d, n) # point update 

    # print(trace_vector)
    # print(C)
    # print(trace_vector)
    C_max, solution = reconstruct_solution(mems, C, trace_vector)

    all_C_max_indicies = all_solutions_c_max_indicies(C,C_max)
    # print(C, C_max)
    print("number solutions with the same score:", all_solutions_c_max_indicies(C, C_max))
    C_max2, solutions = reconstruct_all_solutions(mems, all_C_max_indicies, trace_vector)
    for sol in solutions:
        print(C_max2, [mem.j for mem in sol])

    time_find = time()- st
    # print(C)
    # print(trace_vector)
    # print(C_max, [mem.j for mem in solution])

    # print([ m2.d >= m1.d for m1, m2 in zip(solution[:-1], solution[1:]) ])
    # print([ m2.y >= m1.y for m1, m2 in zip(solution[:-1], solution[1:]) ])
    # assert all( m2.d >= m1.d for m1, m2 in zip(solution[:-1], solution[1:]) )
    # assert all( m2.y >= m1.y for m1, m2 in zip(solution[:-1], solution[1:]) )
    print("Time find nlogn solution:", time_find)
    print("total nlog n", time_construct + time_find)
    for i in range(len(solutions)):
        assert [mem.j for mem in solutions[i]] == [mem.j for mem in solution_quadratic[i]]
    # print("time querying RQ method 2:", time()- st)  








import sys
import re
import math
from itertools import groupby, zip_longest
from array import array
from struct import *
# import parasail
import edlib

from collections import namedtuple, defaultdict

from modules import colinear_solver 
from modules import help_functions


mam = namedtuple('Mam', ['x', 'y', 'c', 'd', 'val', 'j', "min_segment_length", "mam_id", "ref_chr_id"])
globals()[mam.__name__] = mam # Global needed for multiprocessing


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def cigar_to_seq(cigar, query, ref):
    cigar_tuples = []
    result = re.split(r'[=DXSMI]+', cigar)
    i = 0
    for length in result[:-1]:
        i += len(length)
        type_ = cigar[i]
        i += 1
        cigar_tuples.append((int(length), type_ ))

    r_index = 0
    q_index = 0
    q_aln = []
    r_aln = []
    for length_ , type_ in cigar_tuples:
        if type_ == "=" or type_ == "X":
            q_aln.append(query[q_index : q_index + length_])
            r_aln.append(ref[r_index : r_index + length_])

            r_index += length_
            q_index += length_
        
        elif  type_ == "I":
            # insertion w.r.t. reference
            r_aln.append('-' * length_)
            q_aln.append(query[q_index: q_index + length_])
            #  only query index change
            q_index += length_

        elif type_ == 'D':
            # deletion w.r.t. reference
            r_aln.append(ref[r_index: r_index + length_])
            q_aln.append('-' * length_)
            #  only ref index change
            r_index += length_
        
        else:
            print("error")
            print(cigar)
            sys.exit()

    return  "".join([s for s in q_aln]), "".join([s for s in r_aln]), cigar_tuples

def cigar_to_accuracy(cigar_string):
    # score = 0
    cigar_tuples = []
    result = re.split(r'[=DXSMI]+', cigar_string)
    i = 0
    for length in result[:-1]:
        i += len(length)
        type_ = cigar_string[i]
        i += 1
        cigar_tuples.append((int(length), type_ ))
    # print(cigar_tuples)
    aln_length = 0
    matches = 0
    for length_ , type_ in cigar_tuples:
        if type_ == "=":
            matches += length_
            aln_length += length_
            # score += length_
        elif type_ == "X":
            # score += length_*0.1
            aln_length += length_
        else:
            aln_length += length_
    return matches / float(aln_length) #,score/float(aln_length)



def edlib_alignment(query, target, mode = "HW", task = 'locations', k=-1):
    result = edlib.align(query, target, task=task, mode=mode, k=k)
    if result['editDistance'] == -1:
        return [0,0], -1, 0
    
    if task == 'locations':
        locations = result['locations']
        ref_start, ref_stop = locations[0][0], locations[0][1]
        accuracy = ((ref_stop - ref_start) - result['editDistance'])/ (ref_stop - ref_start)
    elif task == 'path':
        locations = result['locations']
        ref_start, ref_stop = locations[0][0], locations[0][1]
        cigar_string = result["cigar"]
        accuracy = cigar_to_accuracy(cigar_string)
        # print(accuracy, ( (ref_stop - ref_start) - result['editDistance'])/ (ref_stop - ref_start))
        # print(cigar_string, result['editDistance'], locations, accuracy)
        query_alignment, target_alignment, cigar_tuples = cigar_to_seq(cigar_string, query, target[ref_start: ref_stop+1 ])
        # print(cigar_string)
        # print(query_alignment)
        # print(target_alignment)

    return result['locations'], result['editDistance'], accuracy #, query_alignment, target_alignment



def calc_complessed_score(read_alignment, ref_alignment, m, n):
    """
        Raw score: R = aI +  bX - cO -  dG
        lambda=1.37 and K=0.711
        E=mn2**-S
    """
    states = ['I' if n1 == n2 else 'G' if n1 == '-' or n2 == '-' else 'X' for n1, n2 in zip(read_alignment,ref_alignment) ]
    compressed_profile = [ (element, len(list(i))) for element, i in groupby(states)] 
    print(compressed_profile)
    # return evalue


def calc_evalue(read_alignment, ref_alignment, m, n):
    """
        Raw score: R = aI +  bX - cO -  dG
        lambda=1.37 and K=0.711
        E=mn2**-S
    """
    a, b, c, d = 1, -1, 1,  1
    lambda_=1.37
    K=0.711

    states = ['I' if n1 == n2 else 'G' if n1 == '-' or n2 == '-' else 'X' for n1, n2 in zip(read_alignment,ref_alignment) ]
    I = states.count('I')
    X = states.count('X')
    G = states.count('G')
    O =  len([s1 for s1, s2 in zip(states[:-1],states[1:]) if s1 != s2 and s2 == 'G'])
    if states[0] == 'G': # starts with a gap
        O += 1
    raw_score = a*I +  b*X - c*O -  d*G
    if raw_score < 0:
        raw_score = 0
    bit_score = (lambda_*raw_score - math.log(K) )/ math.log(2)
    evalue = m*n*2**(-bit_score)
    # print(read_alignment)
    # print(ref_alignment)
    print(I,X,G,O)
    print(raw_score, bit_score, evalue)
    return evalue

def contains(sub, pri):
    M, N = len(pri), len(sub)
    i, LAST = 0, M-N+1
    while True:
        try:
            found = pri.index(sub[0], i, LAST) # find first elem in sub
        except ValueError:
            return False
        if pri[found:found+N] == sub:
            return True
        else:
            i = found+1

def is_overlapping(a_start,a_stop, b_start,b_stop):
    return (int(a_start) <= int(b_start) <= int(a_stop) )  or (int(a_start) <= int(b_stop) <= int(a_stop)) or (int(b_start) <= int(a_start) <= int(a_stop) <= int(b_stop) )


def get_unique_exon_and_flank_locations(solution, parts_to_segments):
    wiggle_overlap = 5
    unique_part_locations = []
    segment_hit_locations = []
    flank_hit_locations = []
    # start_part_offset, part_pos_max = 0, 2**32
    # prev_part = ""
    # approximate_hit_locations = { } # { part_id : (ref_start, ref_stop, read_start, read_stop) }
    partial_segment_hit_locations = { }
    partial_flank_hit_locations = { } 
    choord_to_exon_id = {}

    # these two variables are used to check which exos are considered start and stop exons in the read alignment
    # any such exons will be allowed to align with segments of hits (emulating semi-global alignments)
    first_part_stop = 2**32
    last_part_start = 0
    # print(solution)
    for mem in solution:
        ref_chr_id, ref_start, ref_stop =  mem.exon_part_id.split('^')
        ref_chr_id, ref_start, ref_stop = int(ref_chr_id), int(ref_start), int(ref_stop)
        key_tmp = array('L', [ref_chr_id, ref_start, ref_stop])
        key = key_tmp.tobytes()
        if key not in parts_to_segments:  # is a flank
            flank_hit_locations.append((ref_chr_id, ref_start, ref_stop))  
            # if ref_start - wiggle_overlap <= mem.x < mem.y <= ref_stop + wiggle_overlap:
                # print("made wiggle", ref_start, mem.x , mem.y ,ref_stop)
            if (ref_chr_id, ref_start, ref_stop) in partial_flank_hit_locations:
                # print("in segm, updating")
                partial_flank_hit_locations[(ref_chr_id, ref_start, ref_stop)][1] =  mem.y
                partial_flank_hit_locations[(ref_chr_id, ref_start, ref_stop)][3] =  mem.d
            else: 
                # print("in segm")
                partial_flank_hit_locations[(ref_chr_id, ref_start, ref_stop)] = [mem.x, mem.y, mem.c, mem.d]       
        else:
            # print("Is not a flank, ie exon", (ref_chr_id, ref_start, ref_stop), segm_ids)
            segm_ids = parts_to_segments[key]
            # print(parts_to_segments[key])
            # get all exons associated with the part and see if they are hit
            if ref_stop <= first_part_stop:
                first_part_stop = ref_stop
            if ref_start >= last_part_start:
                last_part_start = ref_start

            for chr_id, s_start, s_stop in grouper(segm_ids, 3):
                # exon overlaps with mem
                # chr_id, s_start, s_stop = unpack('LLL', segm_id)
                if is_overlapping(s_start,s_stop, mem.x, mem.y):
                    segm_id = pack('LLL', chr_id, s_start, s_stop)
                    choord_to_exon_id[(ref_chr_id, s_start, s_stop)] = segm_id
                    # print(s_start, s_stop)
                    # print(s_start,s_stop,  mem.x, mem.y )
                    segment_hit_locations.append( (ref_chr_id, s_start, s_stop) )
                    # if s_start - wiggle_overlap <= mem.x < mem.y <= s_stop + wiggle_overlap:
                        # print("Adding", (ref_chr_id, s_start, s_stop) )
                    if (ref_chr_id, s_start, s_stop) in partial_segment_hit_locations:
                        partial_segment_hit_locations[(ref_chr_id, s_start,s_stop)][1] =  mem.y
                        partial_segment_hit_locations[(ref_chr_id, s_start,s_stop)][3] =  mem.d
                        # print("Upd", (mem.y, mem.d) )
                    else: 
                        partial_segment_hit_locations[(ref_chr_id, s_start,s_stop)] = [mem.x, mem.y, mem.c, mem.d]
                        # print("Init", (mem.x, mem.y, mem.c, mem.d) )

    
    # remove duplicates added and sort to get unique ones
    segment_hit_locations = list(set(segment_hit_locations))
    segment_hit_locations.sort(key= lambda x: x[1])
    # print(segment_hit_locations)
    # print(partial_segment_hit_locations)
    # print(flank_hit_locations)
    # print(partial_flank_hit_locations)
    # print(approximate_hit_locations)
    # print("segment_hit_locations", segment_hit_locations)
    # print("unique_part_locations", unique_part_locations)
    return segment_hit_locations, partial_segment_hit_locations, flank_hit_locations, partial_flank_hit_locations, choord_to_exon_id, first_part_stop, last_part_start


def get_unique_segment_and_flank_choordinates(segment_hit_locations, partial_segment_hit_locations, flank_hit_locations, partial_flank_hit_locations, choord_to_exon_id, segment_to_gene, gene_to_small_segments):
    # compress unique exons to only do alignment once 
    unique_segment_choordinates = defaultdict(set)
    unique_segment_choordinates_partial_hits = defaultdict(set)
    unique_flank_choordinates = defaultdict(set)
    unique_flank_choordinates_partial_hits = defaultdict(set)

    # add all small segments here at once
    if segment_hit_locations:
        unique_genes = set(gene_id for (ref_chr_id, s_start, s_stop) in segment_hit_locations for gene_id in segment_to_gene[choord_to_exon_id[(ref_chr_id, s_start, s_stop)]])
        small_segments = set((small_chr_id, small_start,small_stop) for gene_id in unique_genes for (small_chr_id, small_start,small_stop) in grouper(gene_to_small_segments[gene_id], 3))
        ref_chr_id = segment_hit_locations[0][0]
        for chr_id, s_start, s_stop in small_segments:
            # chr_id, s_start, s_stop = unpack('LLL', small_segment_id)
            small_segment_id = pack('LLL', chr_id, s_start, s_stop)
            unique_segment_choordinates[ (chr_id, s_start, s_stop) ].add(small_segment_id)
    # print(sorted(segment_hit_locations))
    for (ref_chr_id, s_start, s_stop) in segment_hit_locations:
        exon_id = choord_to_exon_id[(ref_chr_id, s_start, s_stop)]
        unique_segment_choordinates[ (ref_chr_id, s_start, s_stop) ].add(exon_id)
        
        if (ref_chr_id, s_start, s_stop) in partial_segment_hit_locations:
            segm_ref_start, segm_ref_stop, segm_read_start, segm_read_stop = partial_segment_hit_locations[(ref_chr_id, s_start, s_stop)]
            unique_segment_choordinates_partial_hits[(ref_chr_id, s_start, s_stop) ] =  (ref_chr_id, segm_ref_start, segm_ref_stop, exon_id)

    for (ref_chr_id, ref_start, ref_stop) in flank_hit_locations:
        # print((ref_start, ref_stop), exon_ids)
        # if not exon_ids: # is a flank
        unique_flank_choordinates[ (ref_chr_id, ref_start, ref_stop) ] = set()
        segm_ref_start, segm_ref_stop, segm_read_start, segm_read_stop = partial_flank_hit_locations[(ref_chr_id, ref_start, ref_stop)]
        # case read starts     read:     [ > 0.2*e_len]   ----------------------------...
        # within start exon    exon: --------------------------------
        if (segm_ref_start - ref_start) > 0.05*(ref_stop - ref_start):
            unique_flank_choordinates_partial_hits[(ref_chr_id, ref_start, ref_stop) ] =  (ref_chr_id, segm_ref_start, segm_ref_stop)

        # case read ends       read:  ...----------------------------   [ > 0.2*e_len]   
        # within end exon      exon:                      ---------------------------------
        if (ref_stop - segm_ref_stop ) > 0.05*(ref_stop - ref_start):
            unique_flank_choordinates_partial_hits[(ref_chr_id,  ref_start, ref_stop) ] =  (ref_chr_id, segm_ref_start, segm_ref_stop)
    # print(unique_segment_choordinates)
    # for u in sorted(unique_segment_choordinates):
    #     print('u',u)
    # sys.exit()
    # assert unique_segment_choordinates_old == unique_segment_choordinates_new
    return unique_segment_choordinates, unique_segment_choordinates_partial_hits, unique_flank_choordinates, unique_flank_choordinates_partial_hits


def add_segment_to_mam(read_seq, ref_chr_id, exon_seq, e_start, e_stop, segm_id, mam_instance, min_acc, annot_label):
    if e_stop - e_start >= 5:
        # exon_seq = ref_segment_sequences[ref_chr_id][(e_start, e_stop)]
        # print((e_start, e_stop))
        # print(exon_seq == ref_seq2)
        # assert exon_seq == ref_seq2
        # print(segm_id, e_stop - e_start)
        # align them to the read and get the best approxinate match
        if e_stop - e_start >= 9:
            locations, edit_distance, accuracy = edlib_alignment(exon_seq, read_seq, mode="HW", task = 'path', k = 0.4*min(len(read_seq), len(exon_seq)) ) 
            # print("{0}_{1}_{2}".format(e_start,e_stop, annot_label), locations, edit_distance, accuracy)
            # if 'flank' in segm_id:
            # print(exon_seq)
            if edit_distance >= 0 and accuracy > min_acc:
                # calc_complessed_score(read_alignment, ref_alignment, len(read_seq), len(exon_seq))
                # e_score = calc_evalue(read_alignment, ref_alignment, len(read_seq), len(exon_seq))
                # start, stop = locations[0]
                # if len(locations) > 1:
                #     print("had more", e_stop - e_start, locations)

                # print(accuracy)
                # print((e_start, e_stop), locations, edit_distance, min_segment_length, accuracy, (min_segment_length - edit_distance)/float(min_segment_length), (stop - start + 1)*accuracy, (stop - start + 1 - edit_distance)* accuracy, (min_segment_length - edit_distance)/float(min_segment_length))
                # print(exon_seq)
                # only take unique start locations by taking the highest accuracy from each hit with the same position
                considered_starts = set()
                max_score = 0
                # locations_filtered = set(s for locs in zip(locations[:-1],locations[1:]) for s in locs if locs[0][0] != locs[1][0])
                # locations_filtered.add(locations[0])
                # locations_filtered.add(locations[-1])
                # print(locations)
                # print(locations_filtered)
                # print("LOOOOOOCS")
                for start, stop in locations:
                    min_segment_length = stop - start + 1 #Edlib end location is inclusive
                    acc_approx = max( (min_segment_length - edit_distance)/min_segment_length, accuracy)
                    score = acc_approx*min_segment_length
                    if start in considered_starts:
                        if score <= max_score:
                            continue
                    else:
                        max_score = score
                    # score = accuracy*matches #(min_segment_length - edit_distance) # accuracy*min_segment_length
                    mam_tuple = mam(e_start, e_stop, start, stop, 
                            score, 0 , min_segment_length, str(segm_id) + annot_label, ref_chr_id) 
                    mam_instance.append(mam_tuple)
                    considered_starts.add(start)
                    max_score = score

        
        else: # small exons between 5-9bp needs exact match otherwise too much noise
            locations, edit_distance, accuracy = edlib_alignment(exon_seq, read_seq, mode="HW", task = 'path', k = 0 )
            # print("HEEERE", exon_seq, locations, e_start, e_stop,ref_chr_id)
            if edit_distance == 0:
                # print("perfect matches:",exon_seq, locations)
                score = len(exon_seq)
                # calc_complessed_score(read_alignment, ref_alignment, len(read_seq), len(exon_seq))
                # e_score = calc_evalue(read_alignment, ref_alignment, len(read_seq), len(exon_seq))
                # for segm_id in all_exon_ids: break # only need one of the redundant exon_ids
                # segm_id = all_exon_ids.pop()
                considered_starts = set()
                for start, stop in locations:
                    if start in considered_starts:
                        continue
                    mam_tuple = mam(e_start, e_stop, start, stop, 
                            score, score, 0, str(segm_id) + annot_label, ref_chr_id) 
                    mam_instance.append(mam_tuple)
                    considered_starts.add(start)
    else:
        pass
        # warning_log_file.write("not aligning exons smaller than 5bp: {0}, {1}, {2}, {3}.\n".format(ref_chr_id, e_start, e_stop, ref_segment_sequences[ref_chr_id][(e_start, e_stop)])) # TODO: align these and take all locations

    if  e_stop - e_start >= 0.8*len(read_seq): # read is potentially contained within exon 
        # print()
        # print("aligning read to exon")
        locations, edit_distance, accuracy = edlib_alignment(read_seq, exon_seq, mode="HW", task = 'path', k = 0.4*min(len(read_seq), len(exon_seq)) )
        # print(exon_seq)
        # print((e_start, e_stop), locations, len(exon_seq), len(read_seq), locations,  edit_distance, accuracy)
        # print()
        if edit_distance >= 0:
            # min_segment_length = min( len(exon_seq) ,len(read_seq) )
            # score = min_segment_length - edit_distance #/len(read_seq)
            
            start, stop = locations[0]
            min_segment_length = stop - start + 1 #Edlib end location is inclusive
            score = accuracy*min_segment_length #matches #*(min_segment_length - edit_distance) #accuracy*min_segment_length  #/len(read_seq)
            # print("LOOK:", min_segment_length, edit_distance, score, locations)
            # if e_score < 1.0:
            if accuracy > min_acc: #(min_segment_length -  edit_distance)/float(min_segment_length) > min_acc:
                start, stop = 0, len(read_seq) - 1
                # covered_regions.append((start,stop, score, segm_id, ref_chr_id))
                # for segm_id in all_exon_ids:
                #     mam_tuple = mam(e_start, e_stop, start, stop, 
                #             score, min_segment_length,  segm_id, ref_chr_id)
                #     mam_instance.append(mam_tuple)
                
                # for segm_id in all_exon_ids: break
                # segm_id = all_exon_ids.pop()
                mam_tuple = mam(e_start, e_stop, start, stop, 
                        score, 0, min_segment_length, str(segm_id) + annot_label, ref_chr_id)
                mam_instance.append(mam_tuple)
    


def main(solution, ref_segment_sequences, ref_flank_sequences, parts_to_segments, segment_to_gene, gene_to_small_segments, read_seq, warning_log_file, min_acc, is_tiling_instance = False):
    """
        NOTE: if paramerer task = 'path' is given to edlib_alignment function calls below, it will give exact accuracy of the aligmnent but the program will be ~40% slower to calling task = 'locations'
            Now we are approxmating accuracy by dividing by start and end of the reference coordinates of the alignment. This is not good approw if there is a large instertion
            in the exon w.r.t. the read.
    """
    # chained_parts_seq = []
    # chained_parts_ids = []
    # prev_ref_stop = -1
    # predicted_transcript = []
    # predicted_exons = []
    # covered_regions = []

    segment_hit_locations, partial_segment_hit_locations, flank_hit_locations, partial_flank_hit_locations, choord_to_exon_id, first_part_stop, last_part_start = get_unique_exon_and_flank_locations(solution, parts_to_segments)
    # print()
    # print(segment_hit_locations)
    # print()

    unique_segment_choordinates, unique_segment_choordinates_partial_hits, \
    unique_flank_choordinates, unique_flank_choordinates_partial_hits = get_unique_segment_and_flank_choordinates(segment_hit_locations, partial_segment_hit_locations, flank_hit_locations, partial_flank_hit_locations, \
                                                                                                     choord_to_exon_id, segment_to_gene, gene_to_small_segments)
    # print()
    # print('unique_exon_choordinate segments', unique_segment_choordinates_partial_hits)
    # for t in sorted(unique_segment_choordinates_partial_hits):
    #     print(t)
    # print()
    # sys.exit()

    # unique_exon_segments = get_segments_of_exons(approximate_hit_locations, unique_segment_choordinates)
    # all_potential_hits = unique_segment_choordinates + unique_exon_segments

    # In the chainer solvers, start and end cordinates are always inclusive, i.e. 1,10 means that the mem
    # spans and includes bases 1,2,...,10. In python indexing of strings we would slice out this interval
    # as [1:11], therefore we subtract 1 from the end of the interval before adding it to MAM instance
    mam_instance = []
    for (ref_chr_id, s_start, s_stop), all_segm_ids in sorted(unique_segment_choordinates.items(), key=lambda x: x[0][1]):
        key_tmp = array('L', [ref_chr_id, s_start, s_stop])
        key = key_tmp.tobytes()
        segment_seq = ref_segment_sequences[key]
        segm_id = all_segm_ids.pop()
        # print("Testing full segment", s_start, s_stop, segm_id, segment_seq, key_tmp)
        add_segment_to_mam(read_seq, ref_chr_id, segment_seq, s_start, s_stop, segm_id, mam_instance, min_acc, annot_label = '_full_segment' )


    # add the flanks if any in the solution But they are required to be start and end flanks of the part MEMs and not overlapping any exons (i.e., the exon hits to be considered)
    for (ref_chr_id, f_start, f_stop), _ in sorted(unique_flank_choordinates.items(), key=lambda x: x[0][1]):
        key_tmp = array('L', [ref_chr_id, f_start, f_stop])
        key = key_tmp.tobytes()
        flank_seq = ref_flank_sequences[key]
        flank_id = "flank_{0}_{1}".format(f_start, f_stop)
        # print("Testing full flank:", f_start, f_stop, flank_seq )
        # if (f_stop <= segment_hit_locations[0][1]) or (segment_hit_locations[-1][2] <= f_start): # is start flank
        add_segment_to_mam(read_seq, ref_chr_id, flank_seq, f_start, f_stop, flank_id, mam_instance, min_acc, annot_label = '_full_flank' )


    # Consider partial hits here after all full exons and flanks have been aligned. A segment is tested for 
    # all exons/flanks with start/ end coordinate after the choort of the last valid MAM added!

    # Do not allow partial hits of internal exons yet (ONLY START and END EXON FOR NOW) because these can generate spurious optimal alignments.
    # print("unique_segment_choordinates_partial_hits" ,unique_segment_choordinates_partial_hits)
    # print(segment_hit_locations)
    # print("first segment_hit_locations:", first_part_stop, segment_hit_locations[0][2])
    # print("Last segment_hit_locations:", last_part_start, segment_hit_locations[-1][1])
    if len(mam_instance) > 0:
        first_valid_mam_stop = min([m.y for m in mam_instance])
        last_valid_mam_start = max([m.x for m in mam_instance])
    else:
        first_valid_mam_stop = -1
        last_valid_mam_start = 2**32
    final_first_stop = max(first_part_stop, first_valid_mam_stop)
    final_last_start = min(last_part_start, last_valid_mam_start)
    # print(first_part_stop >= first_valid_mam_stop, "OMG")
    # print(last_part_start <= last_valid_mam_start, "OMG2")
    segm_already_tried = set()
    for (ref_chr_id, e_start, e_stop) in unique_segment_choordinates_partial_hits:
        key_tmp = array('L', [ref_chr_id, e_start, e_stop])
        key = key_tmp.tobytes()
        ref_chr_id, s_start, s_stop, exon_id = unique_segment_choordinates_partial_hits[(ref_chr_id, e_start, e_stop)]
        segment_seq = ref_segment_sequences[key]        
        if e_stop <= final_first_stop: # is start exon
            partial_segm_seq = segment_seq[s_start - e_start:  ]  # We allow only semi global hit towards one end (the upstream end of the read)
            # print()
            if partial_segm_seq not in segm_already_tried and len(partial_segm_seq) > 5:
                # print("testing partial_hit beginning:", exon_id, e_start, e_stop, s_start, s_stop, final_first_stop, partial_segm_seq)
                add_segment_to_mam(read_seq, ref_chr_id, partial_segm_seq, e_start, e_stop, exon_id, mam_instance, min_acc, annot_label = '_partial_segment_start')
                segm_already_tried.add(partial_segm_seq)
        elif final_last_start <= e_start: # is end_exon
            # print(len(segment_seq), s_start,s_stop, e_start, e_stop, len(segment_seq), s_start - e_start, len(segment_seq) - (e_stop - s_stop +1))
            # partial_segm_seq = segment_seq[s_start - e_start: len(segment_seq) - (e_stop - (s_stop + 1)) ]  # segment is MEM coordinated i.e. inclusive, so we subtract one here
            partial_segm_seq = segment_seq[: len(segment_seq) - (e_stop - (s_stop + 1)) ]  # segment is MEM coordinated i.e. inclusive, so we subtract one here, allow semi global hit towards one end (the downstream end of the read)
            # print()
            if partial_segm_seq not in segm_already_tried and len(partial_segm_seq) > 5:
                # print("testing partial_hit end:", exon_id, e_start, e_stop, s_start, s_stop, final_last_start, partial_segm_seq )
                add_segment_to_mam(read_seq, ref_chr_id, partial_segm_seq, e_start, e_stop, exon_id, mam_instance, min_acc, annot_label = '_partial_segment_end' )
                segm_already_tried.add(partial_segm_seq)


    # finally add eventual partial hits of the flanks if any in the solution But they are required not to overlap any exons 
    segm_already_tried = set()
    for (ref_chr_id, f_start, f_stop) in unique_flank_choordinates_partial_hits:
        key_tmp = array('L', [ref_chr_id, f_start, f_stop])
        key = key_tmp.tobytes()

        ref_chr_id, s_start, s_stop = unique_flank_choordinates_partial_hits[(ref_chr_id, f_start, f_stop)]
        flank_seq = ref_flank_sequences[key]
        flank_id = "flank_{0}_{1}".format(f_start, f_stop)

        partial_flank_seq = flank_seq[s_start - f_start:  ]   # segment is MEM coordinated i.e. inclusive, so we subtract one here
        if partial_flank_seq not in segm_already_tried and len(partial_flank_seq) > 5:
            # print("Testing start flank segment:", f_start, s_stop, partial_flank_seq )
            add_segment_to_mam(read_seq, ref_chr_id, partial_flank_seq, f_start, f_stop, flank_id, mam_instance, min_acc, annot_label = '_partial_flank_start')
            segm_already_tried.add(partial_flank_seq)

        partial_flank_seq = flank_seq[: len(flank_seq) - (f_stop - (s_stop + 1)) ]  # segment is MEM coordinated i.e. inclusive, so we subtract one here
        if partial_flank_seq not in segm_already_tried and len(partial_flank_seq) > 5:
            # print("Testing end flank segment:", s_start, f_stop, partial_flank_seq )
            add_segment_to_mam(read_seq, ref_chr_id, partial_flank_seq, f_start, f_stop, flank_id, mam_instance, min_acc, annot_label = '_partial_flank_end')
            segm_already_tried.add(partial_flank_seq)

    # OLD
    # mam_instance_old = list(filter(lambda x: not("_start" in x.mam_id and x.c >= 50) and not("_end" in x.mam_id and x.d <= len(read_seq) - 50), mam_instance))

    # NEW
    # Save only partial segment/flank hits with the outmost choordinates in each instance
    outmost_partial_start = 2**32
    min_segment_end = 2**32
    outmost_partial_end = 0
    max_segment_start = 0
    for x in mam_instance:
        if "_start" in x.mam_id and x.c < outmost_partial_start:
            outmost_partial_start = x.c
        elif "_end" in x.mam_id and x.d > outmost_partial_end:
            outmost_partial_end = x.d

        if x.d < min_segment_end:
            min_segment_end = x.d
        if x.c > max_segment_start:
            max_segment_start = x.c

    outmost_partial_start = min(outmost_partial_start, min_segment_end) # partial start hit not allowed if further in on read than full hit
    outmost_partial_end = max(outmost_partial_end, max_segment_start) # partial end hit not allowed if further in on read than full hit
    mam_instance = list(filter(lambda x: not("_start" in x.mam_id and x.c > outmost_partial_start) and not("_end" in x.mam_id and x.d < outmost_partial_end), mam_instance))

    # print(mam_instance_old == mam_instance, len(read_seq))
    # print(len(mam_instance_old))
    # print(len(mam_instance))

    mam_instance = sorted(mam_instance, key = lambda x: x.y )
    #  sorted_mems = [ mem(x,y,c,d,val,j,e_id) for j, (x, y, c, d, val, e_id) in enumerate(coordinate_sorted_tuples) ]
    mam_instance  = [mam(m.x, m.y, m.c, m.d, m.val, j, m.min_segment_length, m.mam_id, m.ref_chr_id) for j, m in enumerate(mam_instance) ]  # assingn an index j based on the hit choordinate
    ###################################################################################################
    ###################################################################################################
    ###################################################################################################
    # print("MAM INSTANCE", mam_instance)
    if mam_instance:
        if is_tiling_instance and len(mam_instance) > 100:
            # mam_solution_old, value_old, unique_old = colinear_solver.read_coverage_mam_score(mam_instance, overlap_threshold = 5)
            mam_solution, value, unique = colinear_solver.n_logn_read_coverage_mams(mam_instance, overlap_threshold = 5)
            # if mam_solution != mam_solution2:
            #     print("OLD")
            #     for zzz2 in mam_solution:
            #         print(zzz2)
            #     print()
            #     print("NEW")
            #     for zzz2 in mam_solution2:
            #         print(zzz2)
            #     print(value)
            #     print(value2)
        else:
            mam_solution, value, unique = colinear_solver.read_coverage_mam_score(mam_instance, overlap_threshold = 20)
    else:
        return [], -1, tuple()
    # print(mam_solution)
    covered = sum([mam.d-mam.c + 1 for mam in mam_solution])
    if len(mam_solution) > 0:
        non_covered_regions = []
        non_covered_regions.append( mam_solution[0].c )
        if len(mam_solution) > 1:
            for mam1, mam2 in zip(mam_solution[:-1],mam_solution[1:]):
                non_covered_regions.append( mam2.c - mam1.d -1 )
            # non_covered_regions = [mam2.c-mam1.d for mam1, mam2 in zip(mam_solution[:-1],mam_solution[1:])]
        # add beginning and end
        non_covered_regions.append( len(read_seq)  - mam_solution[-1].d )


    else:
        non_covered_regions = []
    # print(non_covered_regions)
    return non_covered_regions, value, mam_solution


